/*---------------------------------------------------------------------------*\

	FILE....: vtcore.cpp
	AUTHOR..: Ben Kramer, Ron Lee
	DATE....: 01/12/2005
	DESC....: Main part of the vtcore driver for vtcore kernel module cards

\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\

         Voicetronix Voice Processing Board (VPB) Software

         Copyright (C) 1999-2004, 2006 Voicetronix www.voicetronix.com.au

         This library is free software; you can redistribute it and/or
         modify it under the terms of the GNU Lesser General Public
         License as published by the Free Software Foundation; either
         version 2.1 of the License, or (at your option) any later version.

         This library is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty of
         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
         Lesser General Public License for more details.

         You should have received a copy of the GNU Lesser General Public
         License along with this library; if not, write to the Free Software
         Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
         MA  02110-1301  USA

\*---------------------------------------------------------------------------*/

#include "utilfuncs.h"
#include "vtcore.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <string.h>



extern "C" {
#include "alawmulaw.h"
#include "dtmf.h"
#include "toneg.h"
#include "toned.h"
}

#include "mess.h"
#include "messages.h"
#include "generic.h"
#include "dspfifo.h"
#include "mapdev.h"
#include "comm.h"
#include "llconf.h"
#include "vpbdial.h"
#include "vtcore_ioctl.h"

#include "hostvox.h"
#include "v4logagc.h"


/* VPB Driver stuff */
#define ALAW_CODES	256
#define MAX_BOARDS	12
#define FRAME 		8
#define PRI_PORTS	32
//#define NBUF		160
//#define NBUF		80
//#define NBUF		40
//#define NBUF		16
#define AUDIO_SLEEP	20
#define NBUF		(8*AUDIO_SLEEP*2)
//#define NBUF		8*AUDIO_SLEEP*2
//#define NBUF		8*AUDIO_SLEEP
//#define NBUF		16*AUDIO_SLEEP
//#define NBUF		32*AUDIO_SLEEP
#define TDLOOP		(NBUF/40)
#define HOSTDSP_SLEEP	5

#define MAX_BALS	10


// Control functions
static void vtcoreOpenPorts(VPBREG * v);
static void vtcoreSimulateDSP(void *d);
static void vtcoreProcessCommands(VTCore * v);
static void CheckMsgQ(void *d);
static void vtcoreClosePorts(VPBREG * v);

// Audio functions
static void PlayThread(void *pv);
static void RecordThread(void *pv);
static void prepare_rec_buf(unsigned short buf[NBUF],  VTCore * pThis, int ch);
static int prepare_play_buf(unsigned short buf[NBUF],  VTCore * pThis, int ch);
static void toned_callback(word mess[],void *board);

// Configure function
static void vtcoreConfigure(VPBREG * v);
static void vtcore_config_bals(VPBREG *v, int ch, int *bals);
static void vtcore_config_logging(VPBREG *v, int ch, int logging);

// IOCTL helpers
static int vtcore_offhook(int fd, int board, int ch);
static int vtcore_onhook(int fd, int board, int ch);
//static int vtcore_reg_write(int fd, int board, int ch, int reg, int value);
static int vtcore_record_audio(int fd, int board, int ch, unsigned short *buf, int length);
static int vtcore_play_audio(int fd, int board, int ch, unsigned short *buf, int length);
/*
static int vtcore_bridgeon(int fd, int cardnum, word m[COMM_MAX_MESSPCDSP]);
static int vtcore_bridgeoff(int fd, int cardnum, word m[COMM_MAX_MESSPCDSP]);
*/
static int vtcore_listenoff(int fd, int cardnum, word m[COMM_MAX_MESSPCDSP]);
static int vtcore_listenon(int fd, int cardnum, word m[COMM_MAX_MESSPCDSP]);
//static int vtcore_record_fifo_how_full(int fd, int board, int ch);
static int vtcore_play_fifo_how_empty(int fd, int board, int ch);
//static int vtcore_channel_count(int fd, int board);
static void vtcore_set_rec_gain(VPBREG *v, int ch, int gain);
static void vtcore_set_play_gain(VPBREG *v, int ch, int gain);
static int vtcore_echocan_on(int fd, int cardnum, int ch);
static int vtcore_echocan_off(int fd, int cardnum, int ch);
//static int vtcore_port_rst(int fd, int cardnum, int ch);

// Helper functions
//static int sleepms(unsigned int ms);
static int vtcore_get_interface_type(int board, int ch);
static void vtcore_set_ringing(int board, int ch, int rings);

static int pci_fd;
static char pcidev[]="/dev/vt0";

/* Channel state defines */
#define CH_IDLE		0
#define CH_RING		1
#define CH_PROCEEDING	2
#define CH_RINGING	3
#define CH_SETUP	4
#define CH_NEEDANS	5
#define CH_NEEDHANG	6
#define CH_HUNGEDUP	7
#define CH_NEEDRING	8
#define CH_CONNECTED	10

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreHostDSPOpen
	AUTHOR...: Ben Kramer
	DATE.....: 16/11/04


\*--------------------------------------------------------------------------*/

void VTCore::vtcoreHostDSPOpen()
{
	int   i;
	mprintf("[%d/-] Initialising VTCore Host DSP for %s...\n", vr->cardtypnum, vpb_model_desc(vr->model));
	finish = 0;

	if (pci_fd == 0){
		pci_fd = open(pcidev,O_RDWR);
		if(pci_fd == -1){
			printf("Couldnt open up VTCore device nodule (%s): %s\n",
			       pcidev, strerror(errno));
			exit(1);
		}
	}

	//VTCORE_CARD *cardinfo = (VTCORE_CARD *)vr->cardinfo;

	int numch = vr->numch;
	vr->toneg = new void*[numch];
	vr->toned = new void*[numch];
	vtcore_chan * chans = new vtcore_chan[numch];
	vr->chans = chans;

	mprintf("[%d/-] This card has [%d] channels\n", vr->cardtypnum,vr->numch);
	for(i=0 ; i<numch; i++) {
		toneg_open(&vr->toneg[i], i );
		toned_open(&vr->toned[i], i , &toned_callback, (void *)vr);

		/* setup chans structure */
		chans[i].dtmf_dec = dtmf_init();
		chans[i].vox_states = NULL;
		chans[i].agc_states = NULL;
		chans[i].hand_num = -1;	// set the handle number so that we know who we are
		chans[i].chan_num= i;
		chans[i].state=CH_IDLE;
		chans[i].v = vr;
		chans[i].interface = vtcore_get_interface_type(vr->cardtypnum, i);
		mprintf("[%d/%d] Started VTCore channel ...\n",vr->cardtypnum,i);
	}

	vtcoreOpenPorts(vr);

	/* Configure ports from config file! */
	vtcoreConfigure(vr);
	

	GenericInitializeCriticalSection(&mutex);
	Generic_beginthread(vtcoreSimulateDSP, 0, this);
	Generic_beginthread(PlayThread, 0, this);
	Generic_beginthread(RecordThread, 0, this);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreHostDSPClose
	AUTHOR...: Ben Kramer
	DATE.....: 01/12/05

	Constructor to stop host DSP processing.

\*--------------------------------------------------------------------------*/

void VTCore::vtcoreHostDSPClose()
{
	vtcore_chan *chans   = (vtcore_chan *)vr->chans;
	int          cardnum = vr->cardtypnum;
	int                i = 10;

	mprintf("[%d/-] Closing VTCore Host DSP...\n",cardnum);
	finish = 1;
	while( active_threads && --i ) GenericSleep(500);
	if( i == 0 )
		mprintf("[%d/-] FAILED to stop all threads, %d still active\n",
			cardnum, active_threads);

	mprintf("[%d/-] Closing down ports...\n",cardnum);
	vtcoreClosePorts(vr);
	mprintf("[%d/-] Closing down dtmf and tone detectors...\n",cardnum);
	for(i=0; i < vr->numch; ++i) {
		dtmf_close(chans[i].dtmf_dec);
		toneg_close(vr->toneg[i]);
		toned_close(vr->toned[i]);
	}
	mprintf("[%d/-] Closed down dtmf and tone detectors...\n",cardnum);
	mprintf("[%d/-] Freeing structures...\n",cardnum);
	delete [] vr->toneg;
	vr->toneg = NULL;
	delete [] vr->toned;
	vr->toned = NULL;
	delete [] chans;
	chans = NULL;

	GenericDeleteCriticalSection(&mutex);
	mprintf("[%d/-] Closed VTCore Host DSP!\n",cardnum);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreOpenPorts
	AUTHOR...: Ben Kramer
	DATE.....: 01/12/05

	Initialises VTCore ports.

\*--------------------------------------------------------------------------*/

void vtcoreOpenPorts(VPBREG * v)
{
	int     i;
	int cardnum = v->cardtypnum;

	for(i=0; i<v->numch; i++) {
		//vtcore_port_rst(pci_fd, cardnum, i);
		// init DSP (audio) FIFOs
		v->rxdf[i] = NULL;
		v->txdf[i] = NULL;
		v->rxdf[i] = new DspFifo(DSPFIFO_UP, v->szrxdf[i],0);
		v->txdf[i] = new DspFifo(DSPFIFO_DOWN, v->sztxdf[i],0);
		if((v->rxdf[i])&&(v->txdf[i])){
			mprintf("[%d/%d] DSP Audio FIFOs opened OK\n",cardnum,i);
		}
		else {
			mprintf("[%d/%d] DSP Audio FIFOs opened FAILED!\n",cardnum,i);
			assert(0);
		}
	}

}

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreClosePorts
	AUTHOR...: Ben Kramer
	DATE.....: 01/12/05

	Shuts down VTCore ports.

\*--------------------------------------------------------------------------*/

void vtcoreClosePorts(VPBREG * v)
{
	int     i;
	int cardnum = v->cardtypnum;

	for(i=0; i<v->numch; i++) {
		if (v->rxdf[i]){
			delete v->rxdf[i];
			v->rxdf[i]=NULL;
		}
		if (v->txdf[i]){
			delete v->txdf[i]; 
			v->txdf[i]=NULL; 
		}
		mprintf("[%d/%d] DSP Audio FIFOs closed OK\n",cardnum,i);
	}
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreConfigure
	AUTHOR...: Ben Kramer
	DATE.....: 20/12/05

	Applies configuration options from the config file

\*--------------------------------------------------------------------------*/
static void vtcoreConfigure(VPBREG * v)
{
	VTCORE_CARD *card = (VTCORE_CARD*)v->cardinfo;
	llc_var *var = NULL;
	int i,tmp;
	int bals[MAX_BALS];
	int got_bals=0;
	int dtmfms=0;
	int cutthrough=0;
	int logging=0;
	int got_logging=0;

	mprintf("vtcoreConfig: Checking for extra config options...\n");
	var =  (llc_var*)card->config;
	for(i=0;i<MAX_BALS;i++)
		bals[i]=-1;
	while (var){
		if (strncasecmp(var->name,"bal",3)==0){
			got_bals++;
			tmp = atoi(&var->name[3]);
			if(tmp < MAX_BALS){
				//bals[tmp] = atoi(var->value);
				sscanf(var->value,"%x",&(bals[tmp]));
				mprintf("vtcoreConfig: got value for bal reg %d of %d(%s)\n", tmp, bals[tmp],var->value);
			}
			else {
				mprintf("vtcoreConfig: To many ballance registers! %d > % d\n", tmp, MAX_BALS);
			}
		}
		else if (strcasecmp(var->name,"dtmfms")==0){
			dtmfms = atoi(var->value);
			mprintf("vtcoreConfig: got value for dtmfms of %d\n", dtmfms);
			vpbdial_change_dtmf_length(dtmfms, dtmfms);
		}
		else if (strcasecmp(var->name,"cutthrough")==0){
			cutthrough = atoi(var->value);
			mprintf("vtcoreConfig: got value for cutthrough of %d\n", cutthrough);
			v->defRecordGainDown = cutthrough;
		}
		else if (strcasecmp(var->name,"logging")==0){
			logging = atoi(var->value);
			mprintf("vtcoreConfig: got value for logging of %d\n", logging);
			got_logging=1;
		}
		else if (strcasecmp(var->name,"chan")==0){
			tmp = atoi(var->value);
			mprintf("vtcoreConfig: configuring channel %d...\n",tmp);
			if(got_bals)
				vtcore_config_bals(v, tmp, bals);
			if(got_logging){
				mprintf("vtcoreConfig: configuring logging for channel %d...\n",tmp);
				vtcore_config_logging(v,tmp,logging);
			}
			//XXX Why are these disabled?
/*
			if(got_hwplay_gain)
				vtcore_config_hwplay_gain(bals, v, ch);
			if(got_hwrecord_gain)
				vtcore_config_hwrecord_gain(bals, v, ch);
			if(got_play_gain)
				vtcore_config_play_gain(bals, v, ch);
			if(got_record_gain)
				vtcore_config_record_gain(bals, v, ch);
*/
		}
		if(var->next)
			var = var->next;
		else
			var = NULL;
	}
	mprintf("vtcoreConfig: Done checking for extra config options...\n");
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: prepare_play_buf
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04

	Reads bufs of play samples, and formats them ready for playing.
	Also iterates tone generation.

\*--------------------------------------------------------------------------*/

int prepare_play_buf(unsigned short buf[NBUF],  VTCore * pThis, int ch)
{
	int	ret,retread=0;
	word	mess[COMM_MAX_MESSDSPPC];
	int	chdev;
//	vtcore_chan *chans;
//	chans = (vtcore_chan *)v->chans;

	static void     **toneg;
	VPBREG	*v;
	if (!pThis){
		return -1;
	}
	v = pThis->vr;
	toneg=pThis->toneg;

	// Read play fifos, add in tone, convert to alaw
	memset(buf, 0, sizeof(buf));
	retread = v->txdf[ch]->DSPRead( (word*)buf, NBUF);
	ret = toneg_play(v->toneg[ch], (short*)buf, NBUF);
	if (ret == -1) {
		mprintf("Finished playing tone on ch[%d] chdev[%d] card[%d]...\n",ch,chdev,v->cardnum);
		mess[0] = DSP_LTONEG;
		mess[1] = DSP_TONEG;
		mess[2] = ch;
		v->upmess->Write(mess, mess[0]);	
	}
	if (ret == 0)
		return retread;
	else
		return VPB_OK;
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: prepare_rec_buf
	AUTHOR...: Ben Kramer
	DATE.....: 01/12/05

	Transfers buffers of recorded samples to the driver FIFOs.  Also
	runs DTMF decoders on each channel.

\*--------------------------------------------------------------------------*/

void prepare_rec_buf(unsigned short buf[NBUF], VTCore * pThis, int ch)
{
	int	digits_ret=0;
	char	key[2];
	word	mess[COMM_MAX_MESSDSPPC];
	int	ret;
	int	keydown=0;
	VPBREG *v;
	int	i;
	unsigned short agcbuf[NBUF];

	if (!pThis){
		return;
	}
	v = pThis->vr;
	vtcore_chan *chans;
	chans = (vtcore_chan *)v->chans;

	if (chans[ch].agc_states){
		v4log_agc_apply(chans[ch].agc_states, (short*)agcbuf, (short*)buf, NBUF);
		memcpy(buf, agcbuf, sizeof(unsigned short)*NBUF);
	}
	/* write it to the buffer */
	v->rxdf[ch]->DSPWrite( (word*)buf, NBUF);

	// DTMF processing
	for(i=0;i<TDLOOP;i++){
		digits_ret = dtmf_decoder(chans[ch].dtmf_dec, key, 2, (short*)&buf[i*(NBUF/TDLOOP)], NBUF/TDLOOP, &keydown);
		if (digits_ret) {
			// it should be impossible to detect > 1 digit in NBUF sample buffer
			//assert(digits_ret == 1);
	//		mprintf("ch = %d bd = %d DTMF[%d].....\n", ch, v->cardnum,key[0]);
			mess[0] = DSP_LDTMF;
			mess[1] = DSP_DTMF;
			mess[2] = ch;
			mess[3] = key[0];
			ret = v->upmess->Write(mess, mess[0]);	
		}
		else if (keydown){
			mess[0] = DSP_LDTMF_DOWN;
			mess[1] = DSP_DTMF_DOWN;
			mess[2] = ch;
			mess[3] = keydown;
			ret = v->upmess->Write(mess, mess[0]);	
		}
		// tone decoder processing
		toned_analyse(v->toned[ch], (short*)&buf[i*(NBUF/TDLOOP)], NBUF/TDLOOP);
	}
	if (chans[ch].vox_states){
		ret = hostvox_analyse(chans[ch].vox_states,(short*)buf,NBUF);
		if (ret == 1){ // Vox OFF
			mess[0] = DSP_LVOXOFF;
			mess[1] = DSP_VOXOFF;
			mess[2] = ch;
			ret = v->upmess->Write(mess, mess[0]);	
			mprintf("ch = %d bd = %d VOX OFF\n", ch, v->cardnum);
		}
		else if (ret ==2){ // Vox ON
			mess[0] = DSP_LVOXON;
			mess[1] = DSP_VOXON;
			mess[2] = ch;
			ret = v->upmess->Write(mess, mess[0]);	
			mprintf("ch = %d bd = %d VOX ON\n", ch, v->cardnum);
		}
	}

}

/*--------------------------------------------------------------------------*\

	FUNCTION.: toned_callback
	AUTHOR...: David Rowe
	DATE.....: 14/1/03

	This function is called by the tone detector when a tone is detected
	or a tone detector log message is generated.

\*--------------------------------------------------------------------------*/

void toned_callback(word mess[],void *board) 
{
	VPBREG *vr = (VPBREG *)board;


	mprintf("Got Tone[%d] from[%d]\n",mess[3],mess[2]);
	//int ch = mess[2];
	//vtcore_chan *chans;
	//chans = (vtcore_chan *)vr->chans;
	vr->upmess->Write(mess, mess[0]);		
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreSimulateDSP
	AUTHOR...: Ben Kramer
	DATE.....: 01/12/05

	Simulates the DSP on host processing cards:
	- polls card for signalling events
	- audio signal I/O is handled by different threads

\*--------------------------------------------------------------------------*/

void vtcoreSimulateDSP(void *d) 
{
	VTCore* pThis = NULL;
	pThis = (VTCore *)d;

	mprintf("[%d/-] Starting VTCore SimulateDSP\n",pThis->vr->cardnum);
	GenericEnterCriticalSection(&pThis->mutex);
	++pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	while(!pThis->finish) {
		vtcoreProcessCommands(pThis);
		/* Check msg fifo from card */
		CheckMsgQ(pThis);
		GenericSleep(HOSTDSP_SLEEP);		
	}

	GenericEnterCriticalSection(&pThis->mutex);
	--pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	// Clean up thread
	pthread_detach(pthread_self());

	mprintf("[%d/-] Finishing VTCore SimulateDSP\n",pThis->vr->cardnum);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: vtcoreProcessCommands
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04

	Reads commands sent to "DSP" and processes them.

\*--------------------------------------------------------------------------*/

void vtcoreProcessCommands(VTCore * pThis) 
{
	int	read;
//	int	config, lsf_ms, id; 
	int	ch;
	word	mess[COMM_MAX_MESSPCDSP];
	//word	rmess[COMM_MAX_MESSDSPPC];
	int	ret;
	//int	fd = pci_fd;
	void     **toneg;
	VPBREG *v;
	if (pThis != NULL){
		toneg=pThis->toneg;
		v = pThis->vr;
	}
	else {
		mprintf("vtcoreProcessCommands: no pThis, cant continue!\n");
		return;
	}

	vtcore_chan *chans;
	chans = (vtcore_chan *)v->chans;

//	VTCORE_CARD *cardinfo = (VTCORE_CARD *)v->cardinfo;

	int cardnum = v->cardtypnum;

	read = v->dnmess->Read(&mess[0], 1);	
	if (read == OK) {
		read = v->dnmess->Read(&mess[1], mess[0]-1);	
		if (read != OK){
			mprintf("[%d/-]Error on second read of fifo!\n",cardnum);
		}
		ch = mess[2];
		//mprintf("message bytes[%d]: on ch[%d]\n",mess[0],mess[2]);
		switch(mess[1]) {
		case PC_CODEC_ONHK:
			assert(mess[0] == PC_LCODEC_ONHK);
			mprintf("[%d/%d] On Hook\n",cardnum,chans[ch].chan_num);
			chans[ch].state = CH_IDLE;
			vtcore_onhook(pci_fd, cardnum, ch);
			break;

		case PC_CODEC_OFFHK:
			assert(mess[0] == PC_LCODEC_OFFHK);
			ret = -1;
			mprintf("[%d/%d] Off Hook \n",cardnum,ch);
			vtcore_offhook(pci_fd, cardnum, ch);
			break;
/* API uses Listens instead AKA SBRIDGE
		case PC_SPI_BRIDGE:
			assert(mess[0] == PC_LSPI_BRIDGE);
			mprintf("[%d/%d] Bridge on\n",cardnum,ch);
			vtcore_bridgeon(pci_fd, cardnum, mess);
			break;

		case PC_SPI_BRIDGE_OFF:
			assert(mess[0] == PC_LSPI_BRIDGE_OFF);
			mprintf("[%d/%d] Bridge off\n",cardnum,ch);
			vtcore_bridgeoff(pci_fd,mess);
			break;
*/

		case PC_SPI_SBRIDGE:
			assert(mess[0] == PC_LSPI_SBRIDGE);
			mprintf("[%d/%d] Listen on\n",cardnum,ch);
			vtcore_listenon(pci_fd, cardnum, mess);
			break;

		case PC_SPI_SBRIDGE_OFF:
			assert(mess[0] == PC_LSPI_SBRIDGE_OFF);
			mprintf("[%d/%d] Listen off\n",cardnum,ch);
			vtcore_listenoff(pci_fd, cardnum, mess);
			break;


		case PC_CODEC_RECGAIN:
			assert(mess[0] == PC_LCODEC_RECGAIN);
			// TX gain
			mprintf("[%02d/%02d]Setting TX gain to %d\n",cardnum,mess[2],mess[3]);
			vtcore_set_rec_gain(v, mess[2], mess[3]);
			break;

		case PC_CODEC_PLAYGAIN:
			assert(mess[0] == PC_LCODEC_PLAYGAIN);
			mprintf("[%02d/%02d]Setting RX gain to %d\n",cardnum,mess[2],mess[3]);
			vtcore_set_play_gain(v, mess[2], mess[3]);
			break;

		case PC_TONEG:
			assert(mess[0] == PC_LTONEG);
			pThis = (VTCore *)v->hostdsp;
			toneg=pThis->toneg;
			ch = mess[12];
			mprintf("[%d/%d] Starting Tone generation ch(%d)\n",cardnum,mess[12],ch);
			toneg_start(v->toneg[ch], mess);
			break;

		case PC_TONEG_RESET:
//			mprintf("[%d/%d] Reseting Tone generation\n",cardnum,ch);
			pThis = (VTCore *)v->hostdsp;
			toneg=pThis->toneg;
			toneg_reset(v->toneg[ch], mess);
			break;

		case PC_CODEC_BREAK:
			assert(mess[0] == PC_LCODEC_BREAK);
			mprintf("[%d/%02d] Codec Break, aka hook flash? \n",cardnum,ch);
			mprintf("[%d/-] Not supported yet\n",cardnum);
			break;

		case PC_LOOPBACK_ON:
			assert(mess[0] == PC_LLOOPBACK_ON);
			mprintf("[%d/%02d] LoopBack On!\n",cardnum,ch);
			mprintf("[%d/-] Not supported yet\n",cardnum);
			break;

		case PC_LOOPBACK_OFF:
			assert(mess[0] == PC_LLOOPBACK_OFF);
			mprintf("[%d/%02d] LoopBack Off!\n",cardnum,ch);
			mprintf("[%d/-] Not supported yet\n",cardnum);
			break;

		case PC_HOSTECHO_ON:
			assert(mess[0] == PC_LHOSTECHO_ON);
			mprintf("[%d/%02d] Echo canceller enabled \n",cardnum,ch);
			//mprintf("[%d/-] Not supported yet\n",cardnum);
			vtcore_echocan_on(pci_fd, cardnum, ch);
			break;

		case PC_HOSTECHO_OFF:
			assert(mess[0] == PC_LHOSTECHO_OFF);
			mprintf("[%d/%02d] Echo canceller disabled \n",cardnum,ch);
			//mprintf("[%d/-] Not supported yet\n",cardnum);
			vtcore_echocan_off(pci_fd, cardnum, ch);
			break;

		case PC_HOSTECHO_OPT:
			assert(mess[0] == PC_LHOSTECHO_OFF);
			/* none at the moment! */
			mprintf("[%d/%02d] Echo canceller - no optimization at the moment \n",cardnum,ch);
			mprintf("[%d/-] Not supported yet\n",cardnum);
			break;

		case PC_TONED_ST:
			// call progress detector messages 
			toned_add_tone(v->toned[ch], mess);
			break;

		case PC_TONED_ST_DEL:
			toned_del_tone(v->toned[ch], mess);
			break;

		case PC_CODEC_GEN:
			// general purpose word sent to codec config reg
			//id = offset + mess[2];
			//printf("set codec reg ch %d  reg 0x%x val 0x%x\n", id, mess[3], mess[4]);
			//hdaports[id].set_codec_reg(mess[3], mess[4]);
			break;

		case PC_TONED_DEBUG:
			assert(mess[0] == PC_LTONED_DEBUG);
			toned_start_logging(v->toned[ch], mess);
			break;

		case PC_VOX_UPDATE:
			assert(mess[0] == PC_LVOX_UPDATE);
			if(chans[ch].vox_states != NULL){
				hostvox_update_levels(chans[ch].vox_states, mess);
			}
			break;

		case PC_RING_ON:
			if(chans[ch].interface != VPB_FXS){
				mprintf("[%d/-]We dont RING_ON on a FXO port\n",cardnum);
				break;
			}
			if (mess[3] > 255 || mess[3] < 1)
				mess[3] = VPB_RING_STATION_DEFAULT;
			vtcore_set_ringing(cardnum, ch, mess[3]);
			break;

		case PC_RING_OFF:
			if(chans[ch].interface != VPB_FXS){
				mprintf("[%d/-]We dont RING_OFF on a FXO port\n",cardnum);
				break;
			}
			vtcore_set_ringing(cardnum, ch, 0);
			break;

		case PC_URING_OFF:
			mprintf("[%d/-]We dont URING_OFF on VTCore\n",cardnum);
			break;

		case PC_URING_ON:
			mprintf("[%d/-]We dont URING_ON on VTCore\n",cardnum);
			break;

		default:
			mprintf("[%d/-]Unknown DSP message [%d]\n",cardnum, mess[1]);
			//assert(0);
			break;
		}
	}
	else if (read != FIFO_EMPTY){
		mprintf("[%d/-]Error reading VTCore message queue!\n",cardnum);
	}
}


VTCore::VTCore(VPBREG *ivr)
	: vr( ivr )
	, active_threads( 0 )
{
	vtcoreHostDSPOpen();
}

VTCore::~VTCore()
{
	vtcoreHostDSPClose();
}

int VTCore::chan_state(int ch)
{
	vtcore_chan *chans;
	chans = (vtcore_chan *)vr->chans;
	return chans[ch].state;
}

int VTCore::chan_type(int ch)
{
	vtcore_chan *chans;
	chans = (vtcore_chan *)vr->chans;
	return chans[ch].interface;
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: PlayThread
	AUTHOR...: Ben Kramer
	DATE.....: 01/12/2005

	Moves audio samples between the kernel mode driver FIFO and the 
	PC FIFOs.  Play side.

\*--------------------------------------------------------------------------*/
void PlayThread(void *pv)
{
	//int		howempty;
	int		ch;
	unsigned short	buf[NBUF]={};
		
	VTCore*	pThis = (VTCore *)pv;
	VPBREG		*vr = pThis->vr;
	int 		chans = vr->numch;
	struct timeval tv,tv2;
	int timediff=0;
	int tmp;
	int cardnum = vr->cardtypnum;


	GenericEnterCriticalSection(&pThis->mutex);
	++pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	mprintf("[%02d] Starting Play Thread\n",cardnum);
	while(!vr->hostdsp){
		GenericSleep(5);
	}
	mprintf("[%02d] Started Play Thread\n",cardnum);

	int fd  = pci_fd ;

	for(ch=0; ch<chans; ch++) {
		vr->txdf[ch]->SetState(1);
	}
	while(!pThis->finish) {
			
		gettimeofday(&tv,NULL);
		for(ch=0; ch<chans; ch++) {
			memset(buf,0,NBUF*sizeof(unsigned short));
			tmp = vtcore_play_fifo_how_empty(fd, cardnum,ch);
			if (tmp > NBUF){
				if(prepare_play_buf(buf,  pThis, ch) == VPB_OK){
					vtcore_play_audio(fd, cardnum, ch, buf, NBUF);
				}
			}
		}
		GenericSleep(AUDIO_SLEEP);
		gettimeofday(&tv2,NULL);
		timediff = (timediff + (tv2.tv_usec - tv.tv_usec))/2;
	}
	for(ch=0; ch<chans; ch++) {
		vr->txdf[ch]->SetState(0);
	}
	
	GenericEnterCriticalSection(&pThis->mutex);
	--pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	// Clean up thread
	pthread_detach(pthread_self());

	mprintf("[%02d] Finishing Play Thread (avg loop time %d)\n",cardnum,timediff);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: RecordThread
	AUTHOR...: David Rowe
	DATE.....: 11/6/02

	Moves audio samples between the kernel mode driver FIFO and the 
	PC FIFOs.  Record side.

\*--------------------------------------------------------------------------*/

void RecordThread(void *d)
{
	unsigned short	buf[NBUF]={};
	int		ch;
	//int		howfull=0;
	//int		b=0;
	int		cardnum=0;

	VTCore	*pThis = (VTCore *)d;
	VPBREG		*vr = pThis->vr;
	cardnum = vr->cardtypnum;
	int 		chans = vr->numch;
	int fd =  pci_fd;

	GenericEnterCriticalSection(&pThis->mutex);
	++pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	mprintf("[%02d] Starting Record Thread\n",cardnum);
	while(!vr->hostdsp){
		GenericSleep(5);
	}
	mprintf("[%02d] Started Record Thread\n",cardnum);

	for(ch=0; ch<chans; ch++) {
		vr->rxdf[ch]->SetState(1);
	}
	while(!pThis->finish) {
		for(ch=0; ch<chans; ch++) {
			memset(buf,0,NBUF*sizeof(short));
			if(!vtcore_record_audio(fd, cardnum, ch, buf, NBUF))
				prepare_rec_buf(buf, pThis, ch);
		}
		GenericSleep(AUDIO_SLEEP);
	}
	for(ch=0; ch<chans; ch++) {
		vr->rxdf[ch]->SetState(0);
	}

	GenericEnterCriticalSection(&pThis->mutex);
	--pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	// Clean up thread
	pthread_detach(pthread_self());

	mprintf("[%02d] Finishing Record Thread\n",cardnum);
}

#if 0
static int vtcore_port_rst(int fd, int cardnum, int ch)
{
	mprintf("[%02d:%02d] Reseting port\n",cardnum, ch);
	VT_DATA mess;
	mess.board = cardnum;
	mess.channel = ch;
	mess.length = 0;
	mess.data = NULL;
	if( -1 == ioctl(fd, VT_IOC_PORT_RST, &mess))
		return -1;
	else {
		return 0;
	}
}

static int vtcore_reg_read(int fd, int cardnum, int ch, int reg, int * value)
{
	VT_DATA mess;
	VT_REG_RW data;
	mess.board = cardnum;
	mess.channel = ch;
	mess.length = sizeof(VT_REG_RW);
	mess.data = (void *)&data;
	data.reg = reg;
	if( -1 == ioctl(fd, VT_IOC_REGREAD, &mess))
		return -1;
	else {
		*value= data.value;
		return 0;
	}
}

static int vtcore_reg_write(int fd, int cardnum, int ch, int reg, int value)
{
	VT_DATA mess;
	VT_REG_RW data;
	mess.board = cardnum;
	mess.channel = ch;
	mess.length = sizeof(VT_REG_RW);
	mess.data = (void *)&data;
	data.reg = reg;
	data.value = value;
	if( -1 == ioctl(fd, VT_IOC_REGWRITE, &mess))
		return -1;
	else
		return 0;
}
#endif

static int vtcore_record_audio(int fd, int board, int port, unsigned short *buf, int samples)
{
	VT_DATA  mess = { board:    board,
			  channel:  port,
			  length:   samples,
			  data:     buf
			};
	return ioctl(fd, VT_IOC_CHANREADAUDIO, &mess) < 0 ? -1 : 0;
}

static int vtcore_play_audio(int fd, int board, int port, unsigned short *buf, int samples)
{
	VT_DATA  mess = { board:    board,
			  channel:  port,
			  length:   samples,
			  data:     buf
			};
	return ioctl(fd, VT_IOC_CHANWRITEAUDIO, &mess) < 0 ? -1 : 0;
}

static int vtcore_play_fifo_how_empty(int fd, int board, int port)
{
	int tmp = 0;
	VT_DATA  mess = { board:    board,
			  channel:  port,
			  length:   sizeof(int),
			  data:     &tmp
			};
	return ioctl(fd, VT_IOC_CHAN_WRITE_AUDIO_FIFO_HOW_EMPTY, &mess) < 0 ? -1 : tmp;
}

#if 0
static int vtcore_record_fifo_how_full(int fd, int cardnum, int ch)
{
	VT_DATA mess;
	int tmp;
	mess.board = cardnum;
	mess.channel = ch;
	mess.length = sizeof(int);
	mess.data = (void*)&tmp;
	if( -1 == ioctl(fd, VT_IOC_CHAN_READ_AUDIO_FIFO_HOW_FULL, &mess))
		return -1;
	else
		return tmp;
}
#endif

static int vtcore_get_interface_type(int board, int port)
{
	int tmp = VPB_UNKNOWN;
	VT_DATA  mess = { board:    board,
			  channel:  port,
			  length:   sizeof(int),
			  data:     &tmp
			};
	if( ioctl(pci_fd, VT_IOC_BOARD_PORT_TYPE, &mess) >= 0 )
		switch( tmp ) {
		    case VT_FXO_AN: return VPB_FXO;
		    case VT_FXS_AN: return VPB_FXS;
		}
	return VPB_UNKNOWN;
}

static void vtcore_set_ringing(int board, int ch, int rings)
{
	VT_DATA  mess = { board:   board,
			  channel: ch,
			  length:  rings,
			  data:    NULL
			};

	if (rings < 0 || rings > 255)
		mprintf("[%02d/%02d] Set ringing failed! Requires 0 <= rings <= 255, but rings is %d\n", board, ch, rings);
	else if( ioctl(pci_fd, VT_IOC_FXS_RING, &mess) < 0 )
		mprintf("[%02d/%02d] Set ringing failed!\n",board, ch);
	else
		mprintf("[%02d/%02d] Set ringing to %d\n",board, ch, rings);
}

#if 0
static int vtcore_channel_count(int fd, int cardnum)
{
	VT_DATA mess;
	int chans;
	mess.board = cardnum;
	mess.channel = 0;
	mess.length = sizeof(int);
	mess.data = (void *)&chans;
	if( -1 == ioctl(fd, VT_IOC_CHANCOUNT, &mess))
		return -1;
	else
		return chans;
}
#endif


static int vtcore_offhook(int fd, int board, int port)
{
	VT_DATA  mess = { board:   board,
			  channel: port,
			  length:  1,
			  data:    NULL
			};
	return ioctl(fd, VT_IOC_HOOK, &mess) < 0 ? -1 : 0;
}

static int vtcore_onhook(int fd, int board, int port)
{
	VT_DATA  mess = { board:   board,
			  channel: port,
			  length:  0,
			  data:    NULL
			};
	return ioctl(fd, VT_IOC_HOOK, &mess) < 0 ? -1 : 0;
}

static int vtcore_echocan_on(int fd, int board, int port)
{
	VT_DATA  mess = { board:   board,
			  channel: port,
			  length:  0,
			  data:    NULL
			};
	return ioctl(fd, VT_IOC_ECHOCAN_ON, &mess) < 0 ? -1 : 0;
}

static int vtcore_echocan_off(int fd, int board, int port)
{
	VT_DATA  mess = { board:   board,
			  channel: port,
			  length:  0,
			  data:    NULL
			};
	return ioctl(fd, VT_IOC_ECHOCAN_OFF, &mess) < 0 ? -1 : 0;
}

/*
static int vtcore_bridgeon(int fd, int cardnum, word m[COMM_MAX_MESSPCDSP])
{
	VT_DATA mess;
	VT_DATA mess2;
	mess.board = cardnum;    
	mess.channel = m[2];    
	mess.length = sizeof(VT_DATA);
	mess.data = &mess2;
	mess2.board = (short)m[3];
	mess2.channel = (short)m[4];
	mess.data = (void *)(int)m[3];
	if( -1 == ioctl(fd, VT_IOC_BRIDGE, &mess))
		return 1;
	else
		return 0;
}

static int vtcore_bridgeoff(int fd, word m[COMM_MAX_MESSPCDSP])
{
	VT_DATA mess;
	mess.channel = m[2];    
	mess.length = sizeof(int);
	mess.data = (void *)(int)m[3];
	if( -1 == ioctl(fd, VT_IOC_UNBRIDGE, &mess))
		return 1;
	else
		return 0;
}
*/

static int vtcore_listenon(int fd, int board, word m[COMM_MAX_MESSPCDSP])
{
	VT_DATA  msg2 = { board:   m[3],
			  channel: m[4],
			  length:  0,
			  data:    NULL
			};
	VT_DATA  mess = { board:   board,
			  channel: m[2],
			  length:  sizeof(VT_DATA),
			  data:    &msg2
			};
	return ioctl(fd, VT_IOC_LISTEN, &mess) < 0 ? -1 : 0;
}

static int vtcore_listenoff(int fd, int board, word m[COMM_MAX_MESSPCDSP])
{
	VT_DATA  mess = { board:   board,
			  channel: m[2],
			  length:  0,
			  data:    NULL
			};
	return ioctl(fd, VT_IOC_UNLISTEN, &mess) < 0 ? -1 : 0;
}

static void vtcore_config_bals(VPBREG *v, int port, int *bals)
{
	VT_REG_RW   bal;
	VT_DATA     mess = { board:   v->cardtypnum,
			     channel: port,
			     length:  sizeof(VT_REG_RW),
			     data:    &bal
			   };
	int i;
	for(i=0;i<MAX_BALS;i++){
		if (bals[i] != -1){
			bal.reg=i;
			bal.value=bals[i];
			ioctl(pci_fd, VT_IOC_CHAN_BALANCE_SET, &mess);
		}
	}
}

static void vtcore_config_logging(VPBREG *v, int port, int logging)
{
	VT_DATA     mess = { board:   v->cardtypnum,
			     channel: port,
			     length:  logging,
			     data:    NULL
			   };
	mprintf("vtcore_config_logging: setting logging to %d\n", logging);
	ioctl(pci_fd, VT_IOC_SET_LOGGING, &mess);

	vtcore_chan * chans = (vtcore_chan*)v->chans;
	if(logging){
		/* open up vox */
		hostvox_open(&chans[port].vox_states,port);
		//v4log_agc_open(&chans[ch].agc_states);
	}
	else if(chans[port].vox_states != NULL){
		/* close up vox */
		hostvox_close(chans[port].vox_states);
		chans[port].vox_states = NULL;
		//v4log_agc_close(chans[ch].agc_states);
		chans[port].agc_states = NULL;
	}
}

static void vtcore_set_play_gain(VPBREG *v, int port, int gain)
{
	VT_DATA  mess = { board:   v->cardtypnum,
			  channel: port,
			  length:  gain,
			  data:    NULL
			};
	ioctl(pci_fd, VT_IOC_SET_PLAYGAIN, &mess);
}

static void vtcore_set_rec_gain(VPBREG *v, int port, int gain)
{
	VT_DATA  mess = { board:   v->cardtypnum,
			  channel: port,
			  length:  gain,
			  data:    NULL
			};
	ioctl(pci_fd, VT_IOC_SET_RECGAIN, &mess);
}

static void CheckMsgQ(void *d)
{
	char    msgq[8] = { 0 };
	word	rmess[DSP_LDROP];
	VTCore *pThis = (VTCore *)d;
	VPBREG *vr=pThis->vr;
	VT_DATA mess = { board:   vr->cardtypnum,
			 channel: 0,
			 length:  sizeof(msgq),
			 data:    msgq
		       };
	while(ioctl(pci_fd, VT_IOC_READ_MSGQ, &mess) == 0) {
		if (msgq[1] == DSP_CODEC_RING){
			mprintf("Ring ON ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LCODEC_RING;
			rmess[1] = DSP_CODEC_RING;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);

		}
		else if (msgq[1] == DSP_RING_OFF){
			mprintf("Ring OFF ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LRING_OFF;
			rmess[1] = DSP_RING_OFF;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_LOOP_ONHOOK){
			mprintf("LOOP ON hook ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LLOOP_ONHOOK;
			rmess[1] = DSP_LOOP_ONHOOK;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_LOOP_OFFHOOK){
			mprintf("LOOP OFF hook ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LLOOP_OFFHOOK;
			rmess[1] = DSP_LOOP_OFFHOOK;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_LOOP_POLARITY){
			mprintf("LOOP POLARITY ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LLOOP_POLARITY;
			rmess[1] = DSP_LOOP_POLARITY;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_CODEC_HKOFF){
			mprintf("Station OffHook ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LCODEC_HKOFF;
			rmess[1] = DSP_CODEC_HKOFF;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_CODEC_HKON){
			mprintf("Station OnHook ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LCODEC_HKON;
			rmess[1] = DSP_CODEC_HKON;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_CODEC_FLASH){
			mprintf("Station FLASH ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LCODEC_FLASH;
			rmess[1] = DSP_CODEC_FLASH;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else if (msgq[1] == DSP_DROP){
			mprintf("Station OnHook ch [%02d]\n",(int)msgq[2]);
			rmess[0] = DSP_LDROP;
			rmess[1] = DSP_DROP;
			rmess[2] = msgq[2];
			vr->upmess->Write(rmess, rmess[0]);
		}
		else {
			mprintf("Unknown message ! [%d]\n",(int)msgq[1]);
		}
	}
}

/*
static int sleepms(unsigned int ms)
{
	long ms_l = ms;
	struct timespec ts;
	ts.tv_sec = ms_l/1000;
	ts.tv_nsec = (ms_l-ts.tv_sec*1000)*1000000l;
	if(nanosleep(&ts, NULL)){
		mprintf("nanosleep failed!\n");
		return -1;
	}
	return 0;
}
*/
