/*---------------------------------------------------------------------------*\
	vtopenpci.c
	OpenPCI Kernel Module for VoiceTronix modular OpenPCI cards
	Author: Ben Kramer, 24 October 2005
	        Ron Lee, 8 Jun 2006

\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\

	Copyright (C) 2005, 2006 Voicetronix www.voicetronix.com.au
	Copyright (C) 2006 Ron Lee <ron@voicetronix.com.au>

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
	MA  02110-1301  USA

\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

/* Conditional debug options */
#define VERBOSE_TIMING 0

/* Driver constants */
#define DRIVER_DESCRIPTION  "VoiceTronix OpenPCI card driver"
#define DRIVER_AUTHOR       "VoiceTronix <support@voicetronix.com.au>"

#define NAME      "vtopenpci"
#define MAX_PORTS 8            // Maximum number of ports on each carrier
#define MAX_CARDS 8            // Maximum number of carriers per host

#define VT_MOD_INFO_SIZE 16    // Maximum size of a module info data field.

#define DEFAULT_COUNTRY  "AUSTRALIA"


#include <linux/init.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#if VERBOSE_TIMING
 #include <asm/msr.h>          // for rdtscll, cputime.h is overkill here.
#endif

#include "vtcommon.h"
#include "vtopenpci.h"
#include "vtmodule.h"
#include "sivoicedaareg.h"
#include "siproslicreg.h"

#include "messages_dsp_pc.h"


/* Compatibility helpers */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
 #include <linux/interrupt.h>
#else
 typedef void irqreturn_t;
 #define IRQ_NONE
 #define IRQ_HANDLED
 #define IRQ_RETVAL(x)
 #define __devexit_p(x) x
 #define pci_register_driver(x) pci_module_init(x)
#endif


#define INOUT 2

#define VT_PCIDMA_BLOCKSIZE (FRAME * INOUT * MAX_PORTS * 2)
#define VT_PCIDMA_MIDDLE    (FRAME * MAX_PORTS - 4)
#define VT_PCIDMA_END       (FRAME * MAX_PORTS * 2 - 4)

#define PCI_DEVICE_ID_VOICETRONIX_OPENPCI   PCI_ANY_ID


/* macro to address internal TJ registers */
#define TREG(addr)      (card->ioaddr + addr)

/* Tiger320 Registers */
#define TJ_CNTL         TREG(0x00)
#define TJ_OPER         TREG(0x01)
#define TJ_AUXC         TREG(0x02)    /* AUX DIR CONTROL */
#define TJ_AUXD         TREG(0x03)    /* AUX DATA READ/WRITE */
#define TJ_MASK0        TREG(0x04)    /* PCI & DMA CONTROLS */
#define TJ_MASK1        TREG(0x05)    /* SET AUX LINES THAT GEN. INTERUPTS */
#define TJ_INTSTAT      TREG(0x06)    /* SOURCE OF INTERUPT */
#define TJ_AUXR         TREG(0x07)    /* AUX PIN STATUS */

#define TJ_DMAWS        TREG(0x08)
#define TJ_DMAWI        TREG(0x0c)
#define TJ_DMAWE        TREG(0x10)
#define TJ_DMAWC        TREG(0x14)
#define TJ_DMARS        TREG(0x18)
#define TJ_DMARI        TREG(0x1c)
#define TJ_DMARE        TREG(0x20)
#define TJ_DMARC        TREG(0x24)

#define TJ_AUXPMEMSK    TREG(0x29)    /* AUX PME# MASK REGISTER */
#define TJ_AUXINTPOL    TREG(0x2A)    /* AUX INT POLARITY CONTROL */

#define TJ_AUXFUNC      TREG(0x2b)
#define TJ_SERCTL       TREG(0x2d)
#define TJ_FSCDELAY     TREG(0x2f)

#define TJ_REGBASE      TREG(0xc0)    /* HOST INTERFACE (STATUS/CONTROL) aka PIB(0) */
//#define TJ_HBDATA     TREG(0xc4)    /* Host Data Read/Write) aka PIB(1) */

#define TJ_RUNFLAGS     0x01          /* control register 'run mode' flags */

/* macro to address "Peripheral Bus Interface" */
#define PIB(addr)       (TJ_REGBASE + addr * 4)


#define HTXF_READY	(inb(PIB(0)) & 0x10)
#define HRXF_READY	(inb(PIB(0)) & 0x20)

#if VERBOSE_TIMING
 #define REPORT_WAIT(n,x)						    \
	 cardinfo(card->cardnum, #n " wait at %d, " #x " = %d", __LINE__, x )
#else
 #define REPORT_WAIT(n,x)
#endif

#define BUSY_WAIT(countvar,cond,delay,iter,failret)			    \
	countvar=0;							    \
	while(cond){							    \
	    udelay(delay);						    \
	    if(++countvar > iter){					    \
		dbginfo(4,card->cardnum,"busy wait failed at %d",__LINE__);\
		return failret;						    \
	    }								    \
	}								    \
	REPORT_WAIT(busy,i)

#define LOCKED_WAIT(countvar,cond,delay,iter,failret)			    \
	countvar=0;							    \
	while(cond){							    \
	    udelay(delay);						    \
	    if(++countvar > iter){					    \
		dbginfo(4,card->cardnum,"busy wait failed at %d",__LINE__);\
		spin_unlock_irqrestore(&card->lock, flags);		    \
		return failret;						    \
	    }								    \
	}								    \
	REPORT_WAIT(locked,i)

#define HTXF_WAIT()		    BUSY_WAIT(i,HTXF_READY,5,50,RET_FAIL)
#define HRXF_WAIT()		    BUSY_WAIT(i,!HRXF_READY,5,200,RET_FAIL)
#define HTXF_WAIT_RET(failret)	    BUSY_WAIT(i,HTXF_READY,5,50,failret)
#define HRXF_WAIT_RET(failret)	    BUSY_WAIT(i,!HRXF_READY,5,100,failret)

#define HTXF_WAIT_LOCKED()	    LOCKED_WAIT(i,HTXF_READY,5,50,RET_FAIL)
#define HRXF_WAIT_LOCKED()	    LOCKED_WAIT(i,!HRXF_READY,5,100,RET_FAIL)
#define HTXF_WAIT_LOCKED_RET(failret) LOCKED_WAIT(i,HTXF_READY,5,50,failret)
#define HRXF_WAIT_LOCKED_RET(failret) LOCKED_WAIT(i,!HRXF_READY,5,100,failret)


/* Values for ProSlic */
#define VBAT	0xC7	/* 74.824/0.376 */
//#define VBAT	0xBA	/* 69.936/0.376 */

// hook modes (from Si doco)
#define HKMODE_OPEN	0
#define HKMODE_FWDACT	1
#define HKMODE_FWDONACT	2
#define HKMODE_TIPOPEN	3
#define HKMODE_RINGING	4
#define HKMODE_REVACT	5
#define HKMODE_REVONACT	6
#define	HKMODE_RINGOPEN	7

// hook states
#define HOOK_ONHOOK       0
#define HOOK_OFFHOOK      1

struct modinfo {
	char type[4];
	int  ports;
	int  rev;
	char serial[9];
	char date[9];
	struct proc_dir_entry *procfs_root;
};

static struct openpci {
	struct pci_dev *dev;

	struct vtboard  board;
	struct channel  chans[MAX_PORTS];

	int             cardnum;
	int             firmware;

	unsigned long   ioaddr;
	dma_addr_t      readdma;
	dma_addr_t      writedma;
	volatile int   *writeblock;
	volatile int   *readblock;

	spinlock_t      lock;

	unsigned int    interrupt_count;
	unsigned int    arm_interrupt_count;
	int             armbout;
	int             armbin;

	int             ppusage;
	int             ppdatamask;

	//XXX Don't enable this until the next iteration.
	//struct work_struct  arm_work;

	struct modinfo  moduleinfo[2];

    #if VERBOSE_TIMING
	unsigned long long time_tj_last;
	unsigned long long time_tj_last_rel;
	unsigned long long time_tj_max;
	unsigned long long delay_tj_max;
	unsigned long long time_arm_max;
    #endif
} *cards[MAX_CARDS];	//XXX Why do we have this?
			//    It is never used except to determine the
			//    board number.  Do we really care?

// You must hold this lock anytime you access or modify the cards[] array.
DEFINE_MUTEX(cards_mutex);

static unsigned char fxo_port_lookup[8] = { 0x0, 0x8, 0x4, 0xc, 0x10, 0x18, 0x14, 0x1c};
static unsigned char fxs_port_lookup[8] = { 0x0, 0x1, 0x2, 0x3, 0x10, 0x11, 0x12, 0x13};
static const char    vtopenpci[]        = "OpenPCI";

static struct pci_device_id openpci_idtable[] = {
	{ PCI_VENDOR_ID_TIGERJET,    PCI_DEVICE_ID_TIGERJET_300,
	  PCI_VENDOR_ID_VOICETRONIX, PCI_DEVICE_ID_VOICETRONIX_OPENPCI,
	  0, 0, (kernel_ulong_t) &vtopenpci },
	{ 0 }
};
MODULE_DEVICE_TABLE(pci, openpci_idtable);

static int __devinit openpci_probe_board(struct pci_dev *pdev, const struct pci_device_id *ent);
static void __devexit openpci_remove_board(struct pci_dev *pdev);

static struct pci_driver vtopenpci_driver = {
	.name     = NAME,
	.probe    = openpci_probe_board,
	.remove   = __devexit_p(openpci_remove_board),
	.suspend  = NULL,
	.resume   = NULL,
	.id_table = openpci_idtable
};

//XXX Use me!
//static struct workqueue_struct *openpci_workqueue; // = NULL

static int  shareirq = 1;
static int  fastinit; // = 0;
static char *country = DEFAULT_COUNTRY;
static char *program; // = NULL
static int  debug;    // = 0;

/* Diagnostic and fail-mode testing parameters
 * Do not use them and expect normal operation */
//XXX make them bitmasks that let us disable individual ports selectively.
static int  nofxo;  //= 0
static int  nofxs;  //= 0

module_param(shareirq, bool, 0444);
module_param(fastinit, bool, 0444);
module_param(country, charp, 0444);
module_param(program, charp, 0444);
module_param(debug, int, 0644);
module_param(nofxo, bool, 0444);
module_param(nofxs, bool, 0444);
MODULE_PARM_DESC(shareirq, "Enable (or disable) irq sharing");
MODULE_PARM_DESC(fastinit, "Disable fxs leakdown testing");
MODULE_PARM_DESC(country, "Set the default country name");
MODULE_PARM_DESC(program, "Select CPLD or ARM programming mode");
MODULE_PARM_DESC(debug, "Set the logging verbosity");
MODULE_PARM_DESC(nofxo, "Fail to probe FXO modules");
MODULE_PARM_DESC(nofxs, "Fail to probe FXS modules");


// Hardcore diagnostics for monitoring line voltage.
// There is an ongoing cost to using it, so disabled by default.
//
//#define DEBUG_LOOP_VOLTAGE 1
#ifdef DEBUG_LOOP_VOLTAGE
 // This param is a 32 bit bitfield where bit 1 << cardnum * 8 << portnum
 // will enable voltage monitoring on that port (fxo only presently)
 static int voltmeter;        // = 0
 module_param(voltmeter, int, 0600);
 MODULE_PARM_DESC(voltmeter, "Enable loop voltage metering");
#endif


/* Convenience macros for logging */
#define cardinfo(cardnum,format,...) info("[%d] " format, cardnum , ## __VA_ARGS__)
#define portinfo(card,port,format,...)                                        \
	info("[%d] %2d: " format, (card)->cardnum, port , ## __VA_ARGS__)
#define cardwarn(card,format,...) warn("[%d] " format, (card)->cardnum , ## __VA_ARGS__)
#define portwarn(card,port,format,...)                                        \
	warn("[%d] %2d: " format, (card)->cardnum, port , ## __VA_ARGS__)
#define cardcrit(cardnum,format,...) crit("[%d] " format, cardnum , ## __VA_ARGS__)
#define portcrit(card,port,format,...)                                        \
	crit("[%d] %2d: " format, (card)->cardnum, port , ## __VA_ARGS__)
#define dbginfo(n,cardnum,format,...)                                         \
	if(debug>=n) info("[%d] " format, cardnum , ## __VA_ARGS__)
#define dbgport(n,card,port,format,...)	if(debug>=n)                          \
	info("[%d] %2d: " format, (card)->cardnum, port , ## __VA_ARGS__)


int __init vtopenpci_init(void)
{ //{{{
	int ret;

	info(DRIVER_DESCRIPTION " " VT_VERSION " for linux " UTS_RELEASE);
	ret = pci_register_driver(&vtopenpci_driver);
	if (ret < 0){
		crit("Failed to register pci driver (%d)", ret);
		return ret;
	}
	//XXX
    #if 0
	openpci_workqueue = create_workqueue( NAME );
	if (!openpci_workqueue){
		crit("Failed to create workqueue");
		pci_unregister_driver(&vtopenpci_driver);
		return -ENOBUFS;
	}
    #endif
	for( ret=0; cards[ret]; ++ret );
	info("module loaded, driver bound to %d cards", ret);
	if(debug) info("using debug level %d", debug);
	return 0;
} //}}}

void __exit vtopenpci_exit(void)
{ //{{{
	//XXX destroy_workqueue(openpci_workqueue);
	pci_unregister_driver(&vtopenpci_driver);
	info("module exit");
} //}}}


/* You must hold the card spinlock to call this function */
static int __ping_arm(struct openpci *card)
{ //{{{
	int i;
	int pong=0;

	while(pong != 0x02){
		outb(0x02, PIB(1)); ++card->armbout; HTXF_WAIT();
		HRXF_WAIT(); pong = inb(PIB(1)); ++card->armbin;
		dbginfo(5,card->cardnum, "ping_arm returned %x", pong);
	}
	while(pong == 0x02){
		// Poke no-ops into the arm while it is still returning data,
		// if 500 usec elapses with no further response from it then
		// the message queue is should be completely cleared.
		outb(0x00, PIB(1)); ++card->armbout; HTXF_WAIT();
		i = 100;
		while( !HRXF_READY && --i ) udelay(5);
		if( i == 0 ) break;
		pong = inb(PIB(1)); ++card->armbin;
		dbginfo(5,card->cardnum, "ping_arm flushing %x", pong);
	}
	return RET_OK;
} //}}}


/* You must hold the card spinlock to call this function */
static int __read_reg_fxo(struct openpci *card, int port, unsigned char reg, unsigned char *value)
{ //{{{
	int i;
	unsigned char portadr = fxo_port_lookup[port];

	if (HRXF_READY){
		*value = inb(PIB(1)); ++card->armbin;
	}
	outb(0x11, PIB(1));    ++card->armbout; HTXF_WAIT();
	outb(0x2, PIB(1));     ++card->armbout; HTXF_WAIT();
	outb(portadr, PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(reg, PIB(1));     ++card->armbout; HTXF_WAIT();

	HRXF_WAIT(); *value = inb(PIB(1)); ++card->armbin;

	return RET_OK;
} //}}}

static int read_reg_fxo(struct openpci *card, int port, unsigned char reg, unsigned char *value)
{ //{{{
	unsigned long flags;

	spin_lock_irqsave(&card->lock, flags);
	if( __read_reg_fxo(card, port, reg, value) ){
		spin_unlock_irqrestore(&card->lock, flags);
		return RET_OK;
	}
	//XXX We do need to recover if a reg write fails
	//XXX this is not a terribly good way to do it, but its the
	//XXX best we can do for now...
	{ int d = debug; debug = 5; __ping_arm(card); debug = d; }
	spin_unlock_irqrestore(&card->lock, flags);
	portcrit(card,port,"FXO reg %d read FAILED", reg);
	return RET_FAIL;
} //}}}

/* You must hold the card spinlock to call this function */
static int __read_reg_fxs(struct openpci *card, int port, unsigned char reg, unsigned char *value)
{ //{{{
	int i;
	unsigned char portadr = fxs_port_lookup[port];

	if (HRXF_READY){
		*value = inb(PIB(1)); ++card->armbin;
	}
	outb(0x13, PIB(1));    ++card->armbout; HTXF_WAIT();
	outb(0x2, PIB(1));     ++card->armbout; HTXF_WAIT();
	outb(portadr, PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(reg, PIB(1));     ++card->armbout; HTXF_WAIT();

	HRXF_WAIT(); *value = inb(PIB(1)); ++card->armbin;

	return RET_OK;
} //}}}

static int read_reg_fxs(struct openpci *card, int port, unsigned char reg, unsigned char *value)
{ //{{{
	unsigned long flags;

	spin_lock_irqsave(&card->lock, flags);
	if( __read_reg_fxs(card, port, reg, value) ) {
		spin_unlock_irqrestore(&card->lock, flags);
		return RET_OK;
	}
	spin_unlock_irqrestore(&card->lock, flags);
	portcrit(card,port,"FXS reg %d read FAILED", reg);
	return RET_FAIL;
} //}}}

/* You must hold the card spinlock to call this function */
static int __write_reg_fxo(struct openpci *card, int port, unsigned char reg, unsigned char value)
{ //{{{
	int i;
	unsigned char portadr = fxo_port_lookup[port];

	outb(0x10, PIB(1));    ++card->armbout; HTXF_WAIT();
	outb(0x3, PIB(1));     ++card->armbout; HTXF_WAIT();
	outb(portadr, PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(reg, PIB(1));     ++card->armbout; HTXF_WAIT();
	outb(value, PIB(1));   ++card->armbout; HTXF_WAIT();

	return RET_OK;
} //}}}

static int write_reg_fxo(struct openpci *card, int port, unsigned char reg, unsigned char value)
{ //{{{
	unsigned long flags;

	spin_lock_irqsave(&card->lock, flags);
	if( __write_reg_fxo(card, port, reg, value) ){
		spin_unlock_irqrestore(&card->lock, flags);
		return RET_OK;
	}
	spin_unlock_irqrestore(&card->lock, flags);
	portcrit(card,port,"FXO reg %d write(%d) failed", reg, value);
	return RET_FAIL;
} //}}}

/* You must hold the card spinlock to call this function */
static int __write_reg_fxs(struct openpci *card, int port, unsigned char reg, unsigned char value)
{ //{{{
	int i;
	unsigned char portadr = fxs_port_lookup[port];

	outb(0x12, PIB(1));    ++card->armbout; HTXF_WAIT();
	outb(0x3, PIB(1));     ++card->armbout; HTXF_WAIT();
	outb(portadr, PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(reg, PIB(1));     ++card->armbout; HTXF_WAIT();
	outb(value, PIB(1));   ++card->armbout; HTXF_WAIT();

	return RET_OK;
} //}}}

static int write_reg_fxs(struct openpci *card, int port, unsigned char reg, unsigned char value)
{ //{{{
	unsigned long flags;

	spin_lock_irqsave(&card->lock, flags);
	if( __write_reg_fxs(card, port, reg, value) ){
		spin_unlock_irqrestore(&card->lock, flags);
		return RET_OK;
	}
	spin_unlock_irqrestore(&card->lock, flags);
	portcrit(card,port,"FXS reg %d write failed", reg);
	return RET_FAIL;
} //}}}

/* You must hold the card spinlock to call this function */
static int __wait_indreg_fxs(struct openpci *card, int port)
{ //{{{
	unsigned char value;
	int count=100;

	while(--count){
		if( __read_reg_fxs(card,port,PS_IND_ADDR_ST,&value) ) {
			if(value) continue;
			dbgport(7,card,port,"__wait_indreg_fxs %d",100-count);
			return RET_OK;
		} else {
			dbgport(3,card,port,
				"failed to read PS_IND_ADDR_ST, retrying...");
		}
	}
	portcrit(card,port,"FAILED to wait for indirect reg write");
	return RET_FAIL;
} //}}}

static int write_indreg_fxs(struct openpci *card, int port, unsigned char reg, unsigned short value)
{ //{{{
	unsigned long flags;

	spin_lock_irqsave(&card->lock, flags);
	if( __wait_indreg_fxs(card,port)
	 && __write_reg_fxs(card, port, PS_IND_DATA_L, (value & 0xFF))
	 && __write_reg_fxs(card, port, PS_IND_DATA_H, ((value & 0xFF00)>>8))
	 && __write_reg_fxs(card, port, PS_IND_ADDR, reg)
	 && __wait_indreg_fxs(card,port))
	{
		spin_unlock_irqrestore(&card->lock, flags);
		return RET_OK;
	}
	spin_unlock_irqrestore(&card->lock, flags);
	portcrit(card,port,"FXS indreg %d write failed", reg);
	return RET_FAIL;
} //}}}

static int read_indreg_fxs(struct openpci *card, int port, unsigned char reg, unsigned short *value)
{ //{{{
	unsigned long flags;
	unsigned char tmp1, tmp2;

	spin_lock_irqsave(&card->lock, flags);
	if( __wait_indreg_fxs(card,port)
	 && __write_reg_fxs(card, port, PS_IND_ADDR, reg)
	 && __wait_indreg_fxs(card,port)
	 && __read_reg_fxs(card,port,PS_IND_DATA_L,&tmp1)
	 && __read_reg_fxs(card,port,PS_IND_DATA_H,&tmp2) )
	{
		*value = (tmp1 & 0xff) | ((tmp2 & 0xff)<<8);
		// Wait until it has finished!	    //XXX ???
		if(__wait_indreg_fxs(card,port)){
			spin_unlock_irqrestore(&card->lock, flags);
			return RET_OK;
		}
	}
	spin_unlock_irqrestore(&card->lock, flags);
	portcrit(card,port,"FXS indreg %d read failed", reg);
	return RET_FAIL;
} //}}}

#if 0
static int ping_arm(struct openpci *card)
{ //{{{
	unsigned long flags;

	// We may hold this lock for quite a while, but since we are flushing
	// the message queue, and never called from a context where we may be
	// servicing interrupts which might want it, that should be ok.
	spin_lock_irqsave(&card->lock, flags);
	__ping_arm( card );
	spin_unlock_irqrestore(&card->lock, flags);
	return RET_OK;
} //}}}
#endif

static int configure_vdaa_country(struct openpci *card, int port, const char *name)
{ //{{{
	unsigned char value;
	int i;

	for (i=0; i < sizeof(country_regs)/sizeof(struct country_reg); ++i){
		if(!strcmp(country_regs[i].country, name)){
			dbgport(1,card,port, "Setting country to %s", name);
			goto part2;
		}
	}
	return -EINVAL;

    part2:
	/* International control 1 - Register 16 */
	value = (country_regs[i].ohs << 6);
	value |= (country_regs[i].rz << 1);
	value |= (country_regs[i].rt << 0);
	if( ! write_reg_fxo(card, port, VDAR_INTC1, value) ){
		cardcrit(card->cardnum,
			 "write VDAR_INTC1=%x failed on port %d", value, port);
		return -EIO;
	}

	/* DC Termination Control - Register 26 */
	value = (country_regs[i].dcv << 6);
	value |= (country_regs[i].mini << 4);
	value |= (country_regs[i].ilim << 1);
	if( ! write_reg_fxo(card, port, VDAR_DCTERMC, value) ){
		cardcrit(card->cardnum,
			 "write VDAR_DCTERMC=%x failed on port %d", value, port);
		return -EIO;
	}

	/* AC Termination Control - Register 30 */
	value = (country_regs[i].acim << 0);
	if( ! write_reg_fxo(card, port, VDAR_ACTERMC, value) ){
		cardcrit(card->cardnum,
			 "write VDAR_ACTERMC=%x failed on port %d", value, port);
		return -EIO;
	}

	/* DAA Control 5 - Register 31 */
	msleep(1);
	if( ! read_reg_fxo(card, port, VDAR_DAAC5, &value) ){
		cardcrit(card->cardnum,
			 "read VDAR_DAAC5 failed on port %d", port);
		return -EIO;
	}
	value = (value & 0xf7) | (country_regs[i].ohs2 << 3);
	value = value | 0x02;
	if( ! write_reg_fxo(card, port, VDAR_DAAC5, value) ){
		cardcrit(card->cardnum,
			 "write VDAR_DAAC5=%x failed on port %d", value, port);
		return -EIO;
	}
	card->chans[port].country = country_regs[i].country;
	return 0;
} //}}}

static int configure_vdaa_tdm(struct openpci *card, int port)
{ //{{{
	int tmp;

    #ifdef DEBUG_VDAA_BASH_CHECK
	int pcmtscl;

	if( ! read_reg_fxo(card, port, VDAR_PCMTSCL, &pcmtscl) ){
		portcrit(card,port,"read VDAR_PCMTSCL failed");
		return RET_FAIL;
	}
    #endif

	/* Set the Receive/Transmit slot on the TDM bus */
	/* On the OpenPCI4/8 card we have two groups of 4 slots */
	if (port < 4) {
		tmp = port * FRAME;
		//cardinfo(card->cardnum, "Set port %d RX-TDM time slot to %d.",port,tmp);
		if( ! write_reg_fxo(card, port, VDAR_PCMRSCL, tmp) ){
			portcrit(card,port,"write VDAR_PCMRSCL failed");
			return RET_FAIL;
		}
		if( ! write_reg_fxo(card, port, VDAR_PCMRSCH, 0x0) ){
			portcrit(card,port,"write VDAR_PCMRSCH failed");
			return RET_FAIL;
		}
		//cardinfo(card->cardnum, "Set port %d TX-TDM time slot to %d.",port,tmp);
		if( ! write_reg_fxo(card, port, VDAR_PCMTSCL, tmp) ){
			portcrit(card,port,"write VDAR_PCMTSCL failed");
			return RET_FAIL;
		}
		if( ! write_reg_fxo(card, port, VDAR_PCMTSCH, 0x0) ){
			portcrit(card,port,"write VDAR_PCMTSCH failed");
			return RET_FAIL;
		}

	    #ifdef DEBUG_VDAA_BASH_CHECK
		if(tmp != pcmtscl){
			portcrit(card,port,
				 "PCMTSCL was bashed! (tmp[%d] != pcmtscl[%d]",
				 tmp,pcmtscl);
		}
	    #endif
	}
	else if ( port < 8 ){
		tmp = (port -4) * FRAME;
		//cardinfo(card->cardnum, "Set port %d TDM time slot to %d.",port,tmp+512);
		if( ! write_reg_fxo(card, port, VDAR_PCMTSCL, tmp) ){
			portcrit(card,port,"write VDAR_PCMTSCL failed");
			return RET_FAIL;
		}
		if( ! write_reg_fxo(card, port, VDAR_PCMTSCH, 0x02) ){
			portcrit(card,port,"write VDAR_PCMTSCH failed");
			return RET_FAIL;
		}
		if( ! write_reg_fxo(card, port, VDAR_PCMRSCL, tmp) ){
			portcrit(card,port,"write VDAR_PCMRSCL failed");
			return RET_FAIL;
		}
		if( ! write_reg_fxo(card, port, VDAR_PCMRSCH, 0x02) ){
			portcrit(card,port,"write VDAR_PCMRSCH failed");
			return RET_FAIL;
		}

	    #ifdef DEBUG_VDAA_BASH_CHECK
		if(tmp != pcmtscl){
			portcrit(card,port,
				 "PCMTSCL was bashed! (tmp[%d] != pcmtscl[%d]",
				 tmp,pcmtscl);
		}
	    #endif
	}
	return RET_OK;
} //}}}

static int configure_vdaa_port(struct openpci *card, int port)
{ //{{{
	unsigned char value;
	int i, count=3;

	do {
		/* Reset the vdaa registers */
		if( ! write_reg_fxo(card, port, VDAR_CTRL1, 0x80) ){
			portcrit(card,port, "failed to write VDAR_CTRL1");
			return RET_FAIL;
		}
		/* Enable the ISO-cap (Line Side) */
		if( ! write_reg_fxo(card, port, VDAR_DAAC2, 0x00) ){
			portcrit(card,port,"failed to write VDAR_DAAC2");
			return RET_FAIL;
		}
		/* Check that Line Side came up by reading its type */
		i = 0;
		do {
			msleep(1);
			if( read_reg_fxo(card, port, VDAR_REV, &value) ){
				if( value & 0xf0 ){
					dbgport(3,card,port,
						"Got VDAR_REV %d on attempt %d",
						value, i);
					goto part2;
				}
			} else {
				portcrit(card,port, "failed to read VDAR_REV");
				return RET_FAIL;
			}
		} while (++i < 10);

		portwarn(card,port,"Line Side failed to start, retrying");

	} while(--count > 0);

	portwarn(card,port,"Line Side FAILED to start, marking as empty");
	//card->chans[port].porttype = VT_PORT_EMPTY;
	return RET_FAIL;

    part2:
	/* Set for Mu-Law */
	if( ! write_reg_fxo(card,port,VDAR_PCMMODE,0x28) ){
		portcrit(card,port,"failed to write VDAR_PCMMODE");
		return RET_FAIL;
	}

	if( ! configure_vdaa_tdm(card, port) ) return RET_FAIL;

	/* Enable on-hook Line montoring for CID etc */
	if( ! write_reg_fxo(card, port, VDAR_DAAC1, 0x08) ){
		portcrit(card,port, "failed to write VDAR_DAAC1");
		return RET_FAIL;
	}

	/* Set Country - default to Australia */
	switch( configure_vdaa_country(card, port, country) ){
	    case 0: break;
	    case -EINVAL:
		portwarn(card,port, "Country '%s' unknown, using default", country);
		if( configure_vdaa_country(card, port, DEFAULT_COUNTRY) == 0 )
			break;
	    default:
		return RET_FAIL;
	}
	return RET_OK;
} //}}}

static int proslic_dcdc_calibration(struct openpci *card, int port)
{ //{{{
	const int     tst_vbat = VBAT - (5 * 376/1000);
	int           tmp = 0;
	unsigned char value = 0;

	//ping_arm(card);
	/* Set DC-DC Convertor PWM Period */
	if( ! write_reg_fxs(card, port, PS_DCDC_PWM_PERIOD, 0xff) ){
		portcrit(card,port,"failed to write PS_DCDC_PWM_PERIOD");
		return RET_FAIL;
	}

	//ping_arm(card);
	/* Set the DC-DC Convertor Switching Delay */
	/* 0x14 is infact the reset/default value */
	if( ! write_reg_fxs(card, port, PS_DCDC_MIN_OFF_T, 0x19) ){
		portcrit(card,port,"failed to write PS_DCDC_MIN_OFF_T");
		return RET_FAIL;
	}

	//ping_arm(card);
	/* Turn on DC-DC Convertor */
	if( ! write_reg_fxs(card, port, PS_PWR_DN_CTRL1, 0x0) ){
		portcrit(card,port,"dcdc cal failed to write PS_PWR_DN_CTRL1");
		return RET_FAIL;
	}

	/* Wait till PS_VBAT_VOLT1 is upto VBAT volts */
	while (++tmp < 50 && value < tst_vbat){
		msleep(1);
		if( ! read_reg_fxs(card, port, PS_VBAT_VOLT1, &value) ){
			portcrit(card,port,"dcdc cal failed to read PS_VBAT_VOLT1");
			return RET_FAIL;
		}
	}
	if (value < tst_vbat){
		portcrit(card,port,"proslic_dcdc_calibration: "
				   "failed to power up!(%d < %d), retried %d times",
			 value * 376/1000, tst_vbat * 376/1000, tmp);
		if( ! write_reg_fxs(card, port, PS_PWR_DN_CTRL1, 1<<4) ){
			portcrit(card,port,"dcdc cal failed to reset PS_PWR_DN_CTRL1");
		}
		return RET_FAIL;
	}
	dbgport(2,card,port, "power up %d >= %d after %d tries",
		value * 376/1000, tst_vbat * 376/1000, tmp);

	/* Start the DC-DC Converter Peak Current Monitor Calibration */
	if( ! write_reg_fxs(card, port, PS_DCDC_MIN_OFF_T, (0x19|(1<<7))) ){
		portcrit(card,port,"failed to calibrate PS_DCDC_MIN_OFF_T");
		return RET_FAIL;
	}
	tmp = 1;
	do {
		msleep(1);
		if( ! read_reg_fxs(card, port, PS_DCDC_MIN_OFF_T, &value) ){
			portcrit(card,port,"failed to read PS_DCDC_MIN_OFF_T");
			return RET_FAIL;
		}
	} while( (value & (1<<7)) && (++tmp < 500) );
	if(tmp == 500){
		portcrit(card,port,
			 "proslic_dcdc_calibration: failed after %d attempts",tmp);
		return RET_FAIL;
	}

	//ping_arm(card);
	dbgport(2,card,port,"dc-dc calibration completed after %d attempts", tmp);
	return RET_OK;
} //}}}

static int proslic_power_leak_test(struct openpci *card, int port)
{ //{{{
	unsigned char vbat;

	if( ! write_reg_fxs(card, port, PS_MODE, HKMODE_OPEN)
	 || ! write_reg_fxs(card, port, PS_PWR_DN_CTRL1, 0x10) ) goto hell;
	msleep(1000);

	if( ! read_reg_fxs(card, port, PS_VBAT_VOLT1, &vbat) )   goto hell;
	if (vbat < 0x4){
		portcrit(card,port,"FAILED proslic leakage test (%d)",vbat);
		return RET_FAIL;
	}
	return RET_OK;

    hell:
	portcrit(card,port,"FAILED to perform proslic leakdown test");
	return RET_FAIL;
} //}}}

static int proslic_slic_calibration(struct openpci *card, int port)
{ //{{{
	const int reg96 = 0x47;
	const int reg97 = 0x1E; // From AN35
      //const int reg97 = 0x18; // From sample program
	int tmp = 0;
	unsigned char value = 1;

	/* PS_CALIB_CTRL_REG2 
		0x1E ==> 0001 1110
		0x01 ==> 0000 0001
	*/
	/* PS_CALIB_CTRL_REG1 
		0x47 ==> 0100 0111
		0x40 ==> 0100 0000
	*/

	// This is the default state, no need to do it explicitly
    #if 0
	/* disable interrupts */
	if( ! write_reg_fxs(card, port, PS_INT_MASK1, 0x0) ){
		portcrit(card,port,"failed to write PS_INT_MASK1");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_INT_MASK2, 0x0) ){
		portcrit(card,port,"failed to write PS_INT_MASK2");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_INT_MASK3, 0x0) ){
		portcrit(card,port,"failed to write PS_INT_MASK3");
		return RET_FAIL;
	}
    #endif

	/* set open */
	if( ! write_reg_fxs(card, port, PS_MODE, HKMODE_OPEN) ){
		portcrit(card,port,"slic calibration failed to write PS_MODE/HKMODE_OPEN");
		return RET_FAIL;
	}

	if( ! write_reg_fxs(card, port, PS_CALIB_CTRL_REG2, reg97) ){
		cardcrit(card->cardnum,
			 "slic calibration failed to write PS_CALIB_CTRL_REG2");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_CALIB_CTRL_REG1, reg96) ){
		cardcrit(card->cardnum,
			 "slic calibration failed to write PS_CALIB_CTRL_REG1");
		return RET_FAIL;
	}
	while( value && ++tmp < 10 ){
		msleep(160);
		if( ! read_reg_fxs(card, port, PS_CALIB_CTRL_REG1, &value) ){
			cardcrit(card->cardnum,
				 "slic calibration failed to read PS_CALIB_CTRL_REG1");
			return RET_FAIL;
		}
	}
	if( tmp == 10 ){
		cardcrit(card->cardnum,"slic calibration FAILED after %d attempts",tmp);
		return RET_FAIL;
	}

	dbgport(2,card,port,"slic calibration completed after %d attempts", tmp);
	return RET_OK;
} //}}}

static int proslic_longitudinal_balance_calibration(struct openpci *card, int port)
{ //{{{
	unsigned char value = 0;
	int tmp = 0;

	// Ensure the line is on hook before performing this calibration.
	// Change to Active mode to check that.
	if( ! write_reg_fxs(card, port, PS_MODE, HKMODE_FWDACT) ){
		portcrit(card, port,
			 "proslic lb cal: failed to write PS_MODE/HKMODE_FWDACT");
		return RET_FAIL;
	}
	if( ! read_reg_fxs(card, port, PS_LP_COND_IND, &value) ){
		portcrit(card,port,"proslic lb cal: failed to read PS_LP_COND_IND");
		return RET_FAIL;
	}
	if(!(value & 0x04)){
		portcrit(card,port,
			 "proslic_longitudinal_balance_calibration: Off-hook!(0x%02x)",
			 value);
		return RET_FAIL;
	}
	// Back to Open mode again for calibration.
	if( ! write_reg_fxs(card, port, PS_MODE, HKMODE_OPEN) ){
		portcrit(card,port,"proslic lb cal: failed to write PS_MODE/HKMODE_OPEN");
		return RET_FAIL;
	}

	/* Turn on common mode calibration error interrupt */
	if( ! write_reg_fxs(card, port, PS_INT_MASK3, 1<<2) ){
		portcrit(card,port,"proslic lb cal: failed to write PS_INT_MASK3");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_CALIB_CTRL_REG2, 0x01) ){
		portcrit(card,port,"proslic lb cal: failed to write PS_CALIB_CTRL_REG2");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_CALIB_CTRL_REG1, 0x40) ){
		portcrit(card,port,"proslic lb cal: failed to write PS_CALIB_CTRL_REG1");
		return RET_FAIL;
	}
	while( value && ++tmp < 20 ){
		msleep(100);
		if( ! read_reg_fxs(card, port, PS_INT_REG3, &value) ){
			portcrit(card,port,
				 "proslic lb cal: failed to read PS_INT_REG3 (2)");
			return RET_FAIL;
		}
		if( value & 0x04 ) {
			portcrit(card,port,
				 "common mode calibration error, interrupt %#x", value);
			return RET_FAIL;
		}
		if( ! read_reg_fxs(card, port, PS_CALIB_CTRL_REG1, &value) ){
			portcrit(card,port,
				 "proslic lb cal: failed to read PS_CALIB_CTRL_REG1");
			return RET_FAIL;
		}
	}
	if( tmp == 20 ){
		portcrit(card,port,"FAILED lb calibration after %d attempts",tmp);
		return RET_FAIL;
	}
	dbgport(2,card,port,"lb calibration succeeded after %d attempts", tmp);
	return RET_OK;
} //}}}

static int proslic_manual_calibration(struct openpci *card, int port)
{ //{{{
	unsigned char value=1;
	int reg98 = 0x1F;
	int reg99 = 0x1F;

	// ARM is always reset before calibration is done,
	// so there should be no need for any of this.
    #if 0
	/* setting some indirect registers, from sample program, not documented! */
	if( ! write_indreg_fxs(card, port, 88, 0)
	 || ! write_indreg_fxs(card, port, 89, 0)
	 || ! write_indreg_fxs(card, port, 90, 0)
	 || ! write_indreg_fxs(card, port, 91, 0)
	 || ! write_indreg_fxs(card, port, 92, 0)
	 || ! write_indreg_fxs(card, port, 93, 0) ) goto hell;

	/* From sample program */
	// This is necessary if the calibration occurs other than at reset time
	if( ! write_reg_fxs(card, port, PS_CALIB_RING_GAIN, 0x10)
	 || ! write_reg_fxs(card, port, PS_CALIB_TIP_GAIN,  0x10)
	 || ! write_reg_fxs(card, port, PS_CALIB_RING_GAIN, reg98) ) goto hell;
    #endif

	if( ! write_reg_fxs(card, port, PS_CALIB_RING_GAIN, reg98) ) goto hell;
	while(value > 0 && reg98 > 0){
		msleep(40);	// Wait at least 40ms here
		if( ! read_reg_fxs(card, port, PS_Q5_CUR, &value) ) goto hell;
		if (value > 0){
			--reg98;
			if( ! write_reg_fxs(card, port, PS_CALIB_RING_GAIN, reg98) ) goto hell;
		}
	}
	if (reg98 == 0){
		portcrit(card,port,
			 "proslic_manual_calibration: Failed to calibrate RING gain (%d)",
			 value);
		return RET_FAIL;
	}
	value=1;
	if( ! write_reg_fxs(card, port, PS_CALIB_TIP_GAIN, reg99) ) goto hell;
	while(value > 0 && reg99 > 0){
		msleep(40);	// Wait at least 40ms here
		if( ! read_reg_fxs(card, port, PS_Q6_CUR, &value) ) goto hell;
		if (value > 0){
			--reg99;
			if( ! write_reg_fxs(card, port, PS_CALIB_TIP_GAIN, reg99) ) goto hell;
		}
	}
	if (reg99 == 0){
		portcrit(card,port,
			 "proslic_manual_calibration: Failed to calibrate TIP gain (%d)",
			 value);
		return RET_FAIL;
	}

	return RET_OK;

    hell:
	portcrit(card,port,"proslic manual calibration failed");
	return RET_FAIL;
} //}}}

#if 0
static void proslic_voltages(struct openpci *card, int port)
{ //{{{
	int tmp;
	read_reg_fxs(card, port, 82, &tmp);
	printk(NAME ":ProSLIC on module %d powered up to -%d volts (%02x) at VBAT1 \n", port, tmp * 376 / 1000, tmp );
	
//	read_reg_fxs(card, port, 83, &tmp);
//	printk("ProSLIC on module %d powered up to -%d volts (%02x) at VBAT2 \n", port, tmp * 376 / 1000, tmp );

	read_reg_fxs(card, port, 80, &tmp);
	printk(NAME ":ProSLIC on module %d poowered up to -%d volts (%02x) at TIP\n", port, tmp * 376 / 1000, tmp);

	read_reg_fxs(card, port, 81, &tmp);
	printk(NAME ":ProSLIC on module %d powered up to -%d volts (%02x) at RING\n", port, tmp * 376 / 1000, tmp);

} //}}}
#endif

static int configure_proslic_country(struct openpci *card, int port, const char *name)
{ //{{{
	int i;

	for (i=0; i < sizeof(ps_country_regs)/sizeof(struct ps_country_reg); ++i){
		if(!strcmp(ps_country_regs[i].country, name)){
			dbgport(1,card,port, "Setting country to %s", name);
			goto part2;
		}
	}
	return -EINVAL;

    part2:

	if( ! write_reg_fxs(card, port, PS_IMPEDANCE, ps_country_regs[i].value) ){
		portcrit(card,port,"failed to write PS_IMPEDANCE");
		return -EIO;
	}
	card->chans[port].country = ps_country_regs[i].country;
	return 0;
} //}}}

static int configure_proslic_port(struct openpci *card, int port)
{ //{{{
	int i = 0;
	unsigned char value;

	//ping_arm(card);
	/* Write the initial indirect registers */
	for(; i < sizeof(indreg_vals)/sizeof(indreg_vals[0]); ++i){
		if( ! write_indreg_fxs(card, port, indreg_vals[i].reg, indreg_vals[i].value) ){
			cardcrit(card->cardnum,
				 "failed to init proslic ind regs on port %d!", port);
			return RET_FAIL;
		}
	}
	for (i=35; i < 40; ++i){
		if( ! write_indreg_fxs(card, port, i, 0x8000) ){
			portcrit(card,port,"failed to configure proslic ind regs");
			return RET_FAIL;
		}
	}
	if( ! write_reg_fxs(card, port, PS_LOOPBACK, 0) ){
		cardcrit(card->cardnum, "failed to disable loopback");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_FEAT_ENHANCE, 0xEB) ){
		portcrit(card,port,"failed to turn on all Rev E Features");
		return RET_FAIL;
	}

	// Taken from sample program, these next two are not specified by AN35
	// Make VBat switch not automatic
	if( ! write_reg_fxs(card, port, PS_AUTO_MAN_CTRL, 0x17) ){
		portcrit(card,port,"failed to write PS_AUTO_MAN_CTRL");
		return RET_FAIL;
	}
	// The above is a safety measure to prevent Q7
	// from accidentally turning on and burning out.
	// It works in combination with the statement below.
	// Pin 34 DCDRV which is used for the battery switch on the
	// Si3211 & Si3212.  Q7 should be set to OFF for si3210
	if( ! write_reg_fxs(card, port, PS_BAT_CTRL, 0x01) ){
		portcrit(card,port,"failed to write PS_BAT_CTRL");
		return RET_FAIL;
	}

	// 0x32 => 75 volts
	// 0x2e => 69 volts
	// 0x21 => 49.5 volts
	if( ! write_reg_fxs(card, port, PS_VBAT_H, 0x32) ){
		portcrit(card,port,"failed to set High Battery Voltage");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_VBAT_L, 0x11) ){
		portcrit(card,port,"failed to set Low Battery Voltage");
		return RET_FAIL;
	}

	/* Perform DC-DC Calibration */
	if( ! proslic_dcdc_calibration(card, port) ) return RET_FAIL;

	/* Perform power leak test, as sugested in sample program.
	   By its very nature, it takes a while to perform, so make it
	   optional for sites that might restart often.  We must perform
	   DC-DC Calibration again to bring power back up when its done. */
	if( fastinit )
		portwarn(card,port, "skipping proslic leak test");
	else if( ! proslic_power_leak_test(card, port)
	      || ! proslic_dcdc_calibration(card, port) ) return RET_FAIL;

	/* Perform SLIC calibration 'off-hook' */
	if( ! proslic_slic_calibration(card, port) ) return RET_FAIL;

	/* Perform Manual calibration */
	if( ! proslic_manual_calibration(card, port) ) return RET_FAIL;

	//proslic_voltages(card,port);
	if( ! proslic_longitudinal_balance_calibration(card, port) )
		return RET_FAIL;

	/* Flush out energy accumulators */
	for(i=88; i < 96; ++i){
		if( ! write_indreg_fxs(card, port, i, 0x0000) ){
			portcrit(card,port,"failed to configure proslic ind regs (2)");
			return RET_FAIL;
		}
	}
	if( ! write_indreg_fxs(card, port, 97, 0x0000) ){
		portcrit(card,port,"failed to configure proslic ind regs (3)");
		return RET_FAIL;
	}
	for(i=193; i < 212; ++i){
		if( ! write_indreg_fxs(card, port, i, 0x0000) ){
			portcrit(card,port,"failed to configure proslic ind regs (4)");
			return RET_FAIL;
		}
	}

	/* clear interrupts */
	for(i=18; i < 21; ++i){
		if( ! write_reg_fxs(card, port, i, 0xFF) ){
			portcrit(card,port,"failed to clear proslic interrupts");
			return RET_FAIL;
		}
	}
	/* enable interrupts */
	for(i=21; i < 24; ++i){
		if( ! write_reg_fxs(card, port, i, 0xFF) ){
			portcrit(card,port,"failed to enable proslic interrupts");
			return RET_FAIL;
		}
	}

	/* Set time slots */
	if (port < 4) {
		value = port * FRAME;
		//cardinfo(card->cardnum, "Set port %d RX-TDM time slot to %d.",port,value);
		if( ! write_reg_fxs(card, port, PS_PCM_TX_L, value) ){
			portcrit(card,port,"failed to write PS_PCM_TX_L");
			return RET_FAIL;
		}
		if( ! write_reg_fxs(card, port, PS_PCM_TX_H, 0x0) ){
			portcrit(card,port,"failed to write PS_PCM_TX_H");
			return RET_FAIL;
		}
		//cardinfo(card->cardnum, "Set port %d TX-TDM time slot to %d.",port,value);
		if( ! write_reg_fxs(card, port, PS_PCM_RX_L, value) ){
			portcrit(card,port,"failed to write PS_PCM_RX_L");
			return RET_FAIL;
		}
		if( ! write_reg_fxs(card, port, PS_PCM_RX_H, 0x0) ){
			portcrit(card,port,"failed to write PS_PCM_RX_H");
			return RET_FAIL;
		}
	}
	else if ( port < 8 ){
		value = (port -4) * FRAME;
		//cardinfo(card->cardnum, "Set port %d TDM time slot to %d.",port,value+512);
		if( ! write_reg_fxs(card, port, PS_PCM_TX_L, value) ){
			portcrit(card,port,"failed to write PS_PCM_TX_L");
			return RET_FAIL;
		}
		if( ! write_reg_fxs(card, port, PS_PCM_TX_H, 0x02) ){
			portcrit(card,port,"failed to write PS_PCM_TX_H");
			return RET_FAIL;
		}
		if( ! write_reg_fxs(card, port, PS_PCM_RX_L, value) ){
			portcrit(card,port,"failed to write PS_PCM_RX_L");
			return RET_FAIL;
		}
		if( ! write_reg_fxs(card, port, PS_PCM_RX_H, 0x02) ){
			portcrit(card,port,"failed to write PS_PCM_RX_H");
			return RET_FAIL;
		}
	}

	/* Turn on the audio and set to mu-law */
	if( ! write_reg_fxs(card, port, PS_PCM_MODE, 0x28) ){
		portcrit(card,port,"failed to write PS_PCM_MODE");
		return RET_FAIL;
	}

	/* Set up audio path */
	if( ! write_reg_fxs(card, port, PS_LOOPBACK, 0x0) ){
		portcrit(card,port,"failed to write PS_LOOPBACK");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_TXRX_PGC, 0x0) ){
		portcrit(card,port,"failed to write PS_TXRX_PGC");
		return RET_FAIL;
	}
	if( ! write_reg_fxs(card, port, PS_HYB_BAL_RL, 0x33) ){
		portcrit(card,port,"failed to write PS_HYB_BAL_RL");
		return RET_FAIL;
	}

	/* Initilize OSC1, OSC2, Ringing, Pulse Metering */
	/* Leaving at defaults */

	/* Write detect thresholds and filters */
	/* Leaving at defaults */

	/* Write DC feed parameters */
	/* Leaving 65,67 at defaults */
	if( ! write_reg_fxs(card, port, PS_BAT_CTRL, 0x01) ){
		portcrit(card,port,"failed to write PS_BAT_CTRL (2)");
		return RET_FAIL;
	}
	/* 20 mA */
	if( ! write_reg_fxs(card, port, PS_OFHK_LP_CUR, 0x00) ){
		portcrit(card,port,"failed to write PS_OFHK_LP_CUR");
		return RET_FAIL;
	}
	/* -48 */
	if( ! write_reg_fxs(card, port, PS_ONHK_VOLT, 0x20) ){
		portcrit(card,port,"failed to write PS_ONHK_VOLT");
		return RET_FAIL;
	}
	/* -6V */
	if( ! write_reg_fxs(card, port, PS_COMM_MODE_VOLT, 0x04) ){
		portcrit(card,port,"failed to write PS_COMM_MODE_VOLT");
		return RET_FAIL;
	}

	/* Re-write Power Coefficients */
	for(i=35; i < 40; ++i){
		if( ! write_indreg_fxs(card, port, i, indreg_vals[i].value) ){
			portcrit(card,port,"failed to write power coeff");
			return RET_FAIL;
		}
	}

	/* Set operation mode */
	/* Forward active */
	if( ! write_reg_fxs(card, port, PS_MODE, HKMODE_FWDACT) ){
		portcrit(card,port,"config proslic failed to write PS_MODE/HKMODE_FWDACT");
		return RET_FAIL;
	}
	//proslic_voltages(card,port);

	/* Set Country - default to Australia */
	switch( configure_proslic_country(card, port, country) ){
	    case 0: return RET_OK;
	    case -EINVAL:
		portwarn(card,port, "Country '%s' unknown, using default", country);
		if( configure_proslic_country(card, port, DEFAULT_COUNTRY) == 0 )
			return RET_OK;
	}
	return RET_FAIL;
} //}}}


static void vdaa_gain(int scaled_gain, unsigned char *maj, unsigned char *min)
{ //{{{
	int gain = (scaled_gain - 0x80)/10;
//	USHORT scaled_gain = (USHORT)(gain*10.0 + 0x80);

	// do the major
	if (gain >  12)      gain =  12;
	if (gain < -15)      gain = -15;
	if (gain >   0)      *maj = (unsigned char)gain;
	else if (gain == 0 ) *maj = 0;
	else {  //if (gain < 0)
			     *maj  = (unsigned char)(-gain);
			     *maj |= 1<<4;
	}
	// do the minor
	gain = (scaled_gain - 0x80) - (gain*10);
	if (gain > 9)        gain =  9;
	if (gain < -9)       gain = -9;
	if(gain > 0)         *min = (unsigned char)gain;
	else if (gain == 0 ) *min = 0;
	else {  //if (gain < 0)
			     *min  = (unsigned char)(-gain);
			     *min |= 1<<4;
	}
} //}}}

static unsigned short proslic_gain(int gain)
{ //{{{
	if( gain < 0 )    gain = 0;
	if( gain > 0xff ) gain = 0xff;
	return gain << 7;
} //}}}

static int set_play_gain(struct openpci *card, int port, int gain)
{ //{{{
	unsigned char maj,min;

	switch( card->chans[port].porttype ){
	    case VT_PORT_VDAA:
		vdaa_gain(gain,&maj,&min);
		if( ! write_reg_fxo(card, port, VDAR_TXGAIN2, maj)
		 || ! write_reg_fxo(card, port, VDAR_TXGAIN3, min) )
			return -EIO;
		dbgport(4,card,port,"set vdaa tx gain to %#x:%#x", maj, min);
		break;

	    case VT_PORT_PROSLIC:
		// Set just the digital gain, but there is an analogue
		// gain stage that we can manilpulate too if required.
		// For SLIC, tx and rx are relative to the handset.
		if( ! write_indreg_fxs(card, port, PSI_RX_DIG_GAIN, proslic_gain(gain)) )
			return -EIO;
		dbgport(4,card,port,"set proslic rx gain to %#x", proslic_gain(gain));
		break;

	    default:
		return -EINVAL;
	}
	return 0;
} //}}}

static int set_rec_gain(struct openpci *card, int port, int gain)
{ //{{{
	unsigned char maj,min;

	switch( card->chans[port].porttype ){
	    case VT_PORT_VDAA:
		vdaa_gain(gain,&maj,&min);
		if( ! write_reg_fxo(card, port, VDAR_RXGAIN2, maj)
		 || ! write_reg_fxo(card, port, VDAR_RXGAIN3, min) )
			return -EIO;
		dbgport(4,card,port,"set vdaa rx gain to %#x:%#x", maj, min);
		break;

	    case VT_PORT_PROSLIC:
		// Set just the digital gain, but there is an analogue
		// gain stage that we can manilpulate too if required.
		// For SLIC, tx and rx are relative to the handset.
		if( ! write_indreg_fxs(card, port, PSI_TX_DIG_GAIN, proslic_gain(gain)) )
			return -EIO;
		dbgport(4,card,port,"set proslic tx gain to %#x", proslic_gain(gain));
		break;

	    default:
		return -EINVAL;
	}
	return 0;
} //}}}


static inline const char *porttype(struct openpci *card, int port)
{ //{{{
	switch( card->chans[port].porttype ) {
	    case VT_PORT_VDAA:    return "VDAA";
	    case VT_PORT_PROSLIC: return "ProSLIC";
	    case VT_PORT_EMPTY:   return "empty port";
	    default:              return "unknown type";
	}
} //}}}

static int set_hook_state(struct openpci *card, int port, int state)
{ //{{{
	switch( card->chans[port].porttype ) {
	    case VT_PORT_VDAA:
		if (state == HOOK_OFFHOOK){
			dbgport(5, card, port, "taking loop offhook");
			if( write_reg_fxo(card, port, 0x05, 0x09) ){
				card->chans[port].state = CH_OFFHOOK;
				//configure_vdaa_tdm(card, port);
				write_reg_fxo(card, port, 0x20, 0x02);  // LED on
			} else  return -EIO;
		} else {
			dbgport(5,card, port, "putting loop onhook");
			if( write_reg_fxo(card, port, 0x05, 0x08) ){
				card->chans[port].state = CH_IDLE;
				write_reg_fxo(card, port, 0x20, 0x03);  // LED off
			} else  return -EIO;
		}
		break;

	    default:
		portwarn(card,port,"cant change %s hook state", porttype(card,port));
		return -EINVAL;
	}
	return 0;
} //}}}


static int arm_monitor(struct openpci *card, int on)
{ //{{{
	unsigned long flags;
	int i;

	spin_lock_irqsave(&card->lock, flags);
	outb( on ? 0x06 : 0x07, PIB(1) ); ++card->armbout; HTXF_WAIT_LOCKED();
	spin_unlock_irqrestore(&card->lock, flags);

	return RET_OK;
} //}}}

static int vtopenpci_ioctl(void *drvdata, int port, unsigned int cmd, void *data, int length)
{ //{{{
	struct openpci *card = drvdata;
	unsigned long   flags;
	int             i;
	int             tmp, tmp2;
	unsigned char   tmpch;

	if( cmd < VT_IOC_PPWDATA && ! card->chans[port].porttype ){
		portcrit(card,port,"not available");
		return -ENODEV;
	}
	switch(cmd){
	    case VT_IOC_HOOK:
		dbgport(3,card,port, "VT_IOC_HOOK %s", length==1 ? "off":"on");
		return set_hook_state(card,port,length == 1 ? HOOK_OFFHOOK
							    : HOOK_ONHOOK );
	    case VT_IOC_SET_RECGAIN:
		dbgport(3,card,port,"VT_IOC_SET_RECGAIN %s %#x",
			porttype(card,port), length );
		return set_rec_gain(card, port, length);

	    case VT_IOC_SET_PLAYGAIN:
		dbgport(3,card,port, "VT_IOC_SET_PLAYGAIN %s %#x",
			porttype(card,port), length );
		return set_play_gain(card, port, length);

	    case VT_IOC_SET_LOGGING:
		dbgport(3,card,port,"VT_IOC_SET_LOGGING %s",
				    length == 1 ? "on" : "off" );
		switch(card->chans[port].porttype){
		    case VT_PORT_VDAA:
			if(length == 1){
				// Mute transmit side (reg15), then:
				// make sure that on-hook monitoring is enabled (reg5)
				// and that the line is on-hook
				if( ! write_reg_fxo(card, port, VDAR_GAIN, 0x80)
				 || ! write_reg_fxo(card, port, VDAR_DAAC1, 0x08))
					return -EIO;

				card->chans[port].state = CH_MONITOR;
				dbgport(5,card,port,"enabled logging mode");
			} else {
				// Un-mute transmit side
				// No need to reset DAAC1, we usually want it
				// set for CID detection in other modes anyway.
				if( ! write_reg_fxo(card, port, VDAR_GAIN, 0x00) )
					return -EIO;
				card->chans[port].state = CH_IDLE;
				dbgport(5,card,port,"disabled logging mode");
			}
			break;

		    default:
			portwarn(card,port,"cant log from a %s port", porttype(card,port));
			return -EINVAL;
		}
		break;

	    case VT_IOC_ARMWR:
		dbgport(4,card,port,"VT_IOC_ARMWR %#x", length);
		spin_lock_irqsave(&card->lock, flags);
		outb(length, PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED_RET(-EIO);
		spin_unlock_irqrestore(&card->lock, flags);
		break;

	    case VT_IOC_ARMR:
		spin_lock_irqsave(&card->lock, flags);
		HRXF_WAIT_LOCKED_RET(-EIO); tmp = inb(PIB(1)); ++card->armbin;
		spin_unlock_irqrestore(&card->lock, flags);
		dbgport(4,card,port,"VT_IOC_ARMR %#x", tmp);
		if(copy_to_user(data,&tmp,sizeof(int))){
			portcrit(card,port,"VT_IOC_ARMR: failed to copy to user");
			return -EFAULT;
		}
		break;

	    case VT_IOC_JTAGWRITE:
		spin_lock_irqsave(&card->lock, flags);
		// Select CPLD or ARM using ..KO_WRITE apply to HCT241 enable pin's
		tmp = card->ppdatamask;
		// Map the Tiger AUXn pin's to Xilinx DLC5 (JTAG Parallel Cable III)
		tmp |= (length & 0x10 ? 0 : 0x80 )  // Reset to Aux7        D4
		    |  (length & 4 ? 0x10 : 0 )     // TMS to Aux4          D2
		    |  (length & 2 ? 0x20 : 0 )     // TCK to AUX5          D1
		    |  (length & 1 ? 0x40 : 0 );    // TDI to AUX6          D0
		dbgport(5,card,port,"VT_IOC_JTAGWRITE %#x %#x", length, tmp);
		outb(tmp, TJ_AUXD);
		tmp = inb(TJ_CNTL); // Apply reset signal as required (for ARM JTag only).
		if (length & 0x10) {
			outb(tmp | 0x01, TJ_CNTL);
			dbgport(6,card,port,"run mode ERST = high");
		} else {
			outb(tmp & 0xfe, TJ_CNTL);
			dbgport(6,card,port,"reset mode ERST = low");
		}
		spin_unlock_irqrestore(&card->lock, flags);
		break;

	    case VT_IOC_JTAGSTATUS:
		spin_lock_irqsave(&card->lock, flags);
		tmpch = (inb(TJ_AUXR) & 0x04 ? 0x10 : 0 );
		spin_unlock_irqrestore(&card->lock, flags);
		dbgport(5,card,port,"VT_IOC_JTAGSTATUS %#x", tmpch);
		if(copy_to_user(data,&tmpch,sizeof(char))) return -EFAULT;
		break;

	    case VT_IOC_PPRELEASE:
		dbgport(4,card,port,"VT_IOC_PPRELEASE");
		card->ppusage = 0;
		break;

	    case VT_IOC_PPCLAIM:
		// We need to lock this atomically if we are serious,
		// but contention for this should be _extremely_ rare
		// so look elsewhere for the premium sporting events.
		dbgport(4,card,port,"VT_IOC_PPCLAIM");
		if(card->ppusage) return -EBUSY;
		card->ppusage = 1;
		break;

	    //XXX Remove the extra copy from this one too
	    case VT_IOC_OPCI_ARM_RST:
		if(copy_from_user(&tmpch,data,sizeof(char)))
			return -EFAULT;

		spin_lock_irqsave(&card->lock, flags);
		tmp2 = inb(TJ_CNTL);
		outb(tmp2 & (tmpch ? 0x01 : 0xfe), TJ_CNTL);
		spin_unlock_irqrestore(&card->lock, flags);

		dbgport(4,card,port,"VT_IOC_OPCI_ARM_RST set ERST %s",
				    tmpch ? "High" : "Low");
		break;

	    //XXX We need to keep these next three for a little longer,
	    //    but don't use them in new code.
	    case VT_IOC_PPWDATA:
		if(copy_from_user(&tmpch,data,sizeof(char))) return -EFAULT;
		dbgport(5,card,port,"VT_IOC_PPWDATA %#x", tmpch);

		// Select CPLD or ARM using ..KO_WRITE apply to HCT241 enable pin's
		tmp = card->ppdatamask;
		// Map the Tiger AUXn pin's to Xilinx DLC5 (JTAG Parallel Cable III)
		tmp |= (tmpch & 0x10 ? 0 : 0x80 )  // Reset to Aux7        D4
		    |  (tmpch & 4 ? 0x10 : 0 )     // TMS to Aux4          D2
		    |  (tmpch & 2 ? 0x20 : 0 )     // TCK to AUX5          D1
		    |  (tmpch & 1 ? 0x40 : 0 );    // TDI to AUX6          D0
		spin_lock_irqsave(&card->lock, flags);
		outb(tmp, TJ_AUXD);
		tmp = inb(TJ_CNTL); // Apply reset signal as required (for ARM JTag only).
		if (tmpch & 0x10) {
			outb(tmp | 0x01, TJ_CNTL);
			dbgport(6,card,port,"run mode ERST = high");
		} else {
			outb(tmp & 0xfe, TJ_CNTL);
			dbgport(6,card,port,"reset mode ERST = low");
		}
		spin_unlock_irqrestore(&card->lock, flags);
		break;

	    case VT_IOC_PPRSTATUS:
		spin_lock_irqsave(&card->lock, flags);
		tmp = (inb(TJ_AUXR) & 0x04 ? 0x10 : 0 );
		spin_unlock_irqrestore(&card->lock, flags);
		dbgport(5,card,port,"VT_IOC_PPRSTATUS %#x", tmp);
		if(copy_to_user(data,&tmp,sizeof(int))) return -EFAULT;
		break;

	    case VT_IOC_ARMW:
		if(copy_from_user(&tmp,data,sizeof(int))){
			portcrit(card,port,"VT_IOC_ARMW: failed to copy from user");
			return -EFAULT;
		}
		dbgport(4,card,port,"VT_IOC_ARMW %#x", tmp);
		spin_lock_irqsave(&card->lock, flags);
		outb(tmp, PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED_RET(-EIO);
		spin_unlock_irqrestore(&card->lock, flags);
		break;

	    case VT_IOC_OPCI_ARM_RESET:
		portcrit(card,port,"VT_IOC_OPCI_ARM_RESET is obsolete");
		return -EINVAL;
		// This ioctl is no longer used and apparently never really
		// worked properly anyway.  Use ping_arm to flush it instead.
	    #if 0
		reset_arm(card);
		if( ! flush_arm(card) ){
			portcrit(card,port,"failed to flush ARM");
			return -EIO;
		}
		cardinfo(card->cardnum, "ARM Reset");
		break;
	    #endif

	    #if 0
	    case VT_IOC_PORT_RST: //{{{
		dbgport(4,card,port,"VT_IOC_PORT_RST");
		if( card->chans[port].porttype == VT_PORT_VDAA
		 && ! configure_vdaa_port(card,port) ){
			portwarn(card,port,"VT_IOC_PORT_RST - failed to reset VDAA port");
			return -EIO;
		}
		//spin_lock_irqsave(&card->lock, flags);
		if(card->firmware > 2){
			/* disable module monitoring on Arm */
			arm_monitor(card, 0);
		}
		disable_interrupts(card);
		if (card->chans[port].porttype == VT_PORT_VDAA){
			if( ! configure_vdaa_port(card,port) ){
				cardinfo(card->cardnum,
					 "%02d: VT_IOC_PORT_RST failed to reset VDAA",
					 port);
				return -EIO;
			}
		}
		enable_interrupts(card);
		if(card->firmware > 2){
			/* Ping Arm to make sure its buffers are clean */
			if( ! ping_arm(card) )
				cardinfo(card->cardnum, "ping_arm failed");
			/* enable module monitoring on Arm */
			arm_monitor(card, 1);
			cardinfo(card->cardnum,
				 "%02d: enabling Arm to monitor ports", port);
		}
		//spin_unlock_irqrestore(&card->lock, flags);
		break;

	    case VT_IOC_REGWRITE:
		VT_REG_RW   reg;
		if( copy_from_user(&reg,(void *)data,sizeof(VT_REG_RW)) ){
			portcrit(card,port,"VT_IOC_REGWRITE copy of VT_REG_RW failed");
			return -EFAULT;
		}
		dbgport(4,card,port,"VT_IOC_REGWRITE %s %x = %x",
			porttype(card,port), reg.reg, reg.value );
		if( card->chans[port].porttype == VT_PORT_VDAA
		 && ! write_reg_fxo(card, port, reg.reg, reg.value) )
			return -EIO;
		if( card->chans[port].porttype == VT_PORT_PROSLIC
		 && ! write_reg_fxs(card, port, reg.reg, reg.value) )
			return -EIO;
		break;

	    case VT_IOC_REGREAD:
		VT_REG_RW   reg;
		if( copy_from_user(&reg,(void *)data,sizeof(VT_REG_RW)) ){
			portcrit(card,port,"VT_IOC_REGREAD copy of VT_REG_RW failed");
			return -EFAULT;
		}
		if( card->chans[port].porttype == VT_PORT_VDAA
		 && ! read_reg_fxo(card, port, reg.reg, &reg.value) ){
			return -EIO;
		} else if( card->chans[port].porttype == VT_PORT_PROSLIC
			&& ! read_reg_fxs(card, port, reg.reg, &reg.value) ){
			return -EIO;
		}
		dbgport(4,card,port,"VT_IOC_REGREAD %s %x = %x",
			porttype(card,port), reg.reg, reg.value );
		if(copy_to_user(data,&reg,sizeof(VT_REG_RW))) return -EFAULT;
		break; //}}}
	    #endif

	    default:
		portcrit(card,port,"unknown ioctl type %d", cmd);
		return -EINVAL;
	}
	return 0;
} //}}}

/* You must hold the card spinlock to call this function */
static int __get_module_info(struct openpci *card, int module, int field, char *value)
{ //{{{
	int i;
	int x=0;
	int count=0;

	*value = 0;
	outb(0x40,   PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(module, PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(field,  PIB(1)); ++card->armbout; HTXF_WAIT();
	HRXF_WAIT(); count = inb(PIB(1)); ++card->armbin;
	if (count > VT_MOD_INFO_SIZE){
		cardcrit(card->cardnum, "Too many bytes of module(%d:%d) data %d/%d",
					 module, field, count, VT_MOD_INFO_SIZE);
		return RET_FAIL;
	}
	dbginfo(4,card->cardnum, "get_module_info(%d:%d): byte count %d",
				 module, field, count);
	for(; x < count; ++x){
		HRXF_WAIT(); *value = inb(PIB(1)); ++card->armbin;
		dbginfo(4,card->cardnum, "get_module_info(%d:%d): byte %d => 0x%02x",
					 module, field, x, *value);
		++value;
	}
	return RET_OK;
} //}}}

static int get_port_type(struct openpci *card, int port)
{ //{{{
	int i, type;
	unsigned long flags;

	spin_lock_irqsave(&card->lock, flags);
	outb(0x20, PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED_RET(VT_PORT_EMPTY);
	outb(port, PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED_RET(VT_PORT_EMPTY);
	HRXF_WAIT_LOCKED_RET(VT_PORT_EMPTY); type = inb(PIB(1)); ++card->armbin;
	spin_unlock_irqrestore(&card->lock, flags);

	return type;
} //}}}

static int check_ports(struct openpci *card)
{ //{{{
	int portcount = 0;
	int i = 0;

	for(; i < MAX_PORTS; ++i ){
		struct channel *chan = &card->chans[i];

		chan->porttype = get_port_type(card, i);
		dbgport(1,card,i,"%s", porttype(card,i));

		// XXX check nofx{o,s} here...
		if( chan->porttype != VT_PORT_EMPTY ){
			chan->codec = VT_MULAW;
			++portcount;
		}
	}
	cardinfo(card->cardnum, "found %d active ports", portcount);
	// we 'succeed' if any ports were discovered.
	return portcount ? RET_OK : RET_FAIL;
} //}}}

// Do not call this from an interrupt context, it may (indirectly) sleep.
static int configure_ports(struct openpci *card)
{ //{{{
	int portcount = 0;
	int i = 0;

	for(; i < MAX_PORTS; ++i){
	    switch( card->chans[i].porttype ){
		case VT_PORT_VDAA:
			if( ! configure_vdaa_port(card,i) ){
				cardcrit(card->cardnum,
					 "FAILED to configure vdaa port %d.  Disabled.", i);
				card->chans[i].porttype = VT_PORT_EMPTY;
			} else
				++portcount;
			break;

		case VT_PORT_PROSLIC:
			if( ! configure_proslic_port(card,i) ){
				cardcrit(card->cardnum,
					 "FAILED to configure proslic port %d.  Disabled.", i);
				card->chans[i].porttype = VT_PORT_EMPTY;
			} else
				++portcount;
			break;

		case VT_PORT_EMPTY:
			break;

		default:
			portcrit(card,i,"unknown port type %d unconfigured",
				 card->chans[i].porttype);
			break;
	    }
	}
	// we 'succeed' if any ports were configured successfully.
	return portcount ? RET_OK : RET_FAIL;
} //}}}

/* You must hold the card spinlock to call this function */
static int __get_arm_id(struct openpci *card, int field, char *value)
{ //{{{
	int i;
	int x=0;
	int count=0;

	outb(0x01,  PIB(1)); ++card->armbout; HTXF_WAIT();
	outb(field, PIB(1)); ++card->armbout; HTXF_WAIT();
	HRXF_WAIT(); count = inb(PIB(1)); ++card->armbin;
	if (count > ID_DATA_MAXSIZE){
		cardcrit(card->cardnum, "Too many bytes of id(%d) data %d/%d",
					 field, count, ID_DATA_MAXSIZE);
		return RET_FAIL;
	}
	dbginfo(4,card->cardnum, "get_arm_id(%d): byte count %d",field,count);
	for(; x < count; ++x){
		HRXF_WAIT(); *value = inb(PIB(1)); ++card->armbin;
		dbginfo(4,card->cardnum, "get_arm_id(%d): byte %d => 0x%02x",field,x,*value);
		++value;
	}
	return RET_OK;
} //}}}

static int check_arm(struct openpci *card)
{ //{{{
	char model[ID_DATA_MAXSIZE+1] = { 0 };
	char date[ID_DATA_MAXSIZE+1]  = { 0 };
	unsigned long flags;
	int i=0, tmp=0;

	spin_lock_irqsave(&card->lock, flags);
	card->firmware = 0;
	while (tmp != 0x88 && ++i < 1000){
		outb(0x88, PIB(0)); ++card->armbout;
		mdelay(1);
		tmp = inb(PIB(1)); ++card->armbin;
	}
	if(i>=1000) goto limbo;
	dbginfo(1,card->cardnum, "Arm responded on attempt %d",i);

	// Flush out the queue if we sent several pings before a response.
	if(i>1)	__ping_arm(card);

	if( ! __get_arm_id(card, 0, model) )  goto hell;
	sscanf(model, "OpenPCI8.%02d", &(card->firmware));
	cardinfo(card->cardnum, "  model: %s", model);

	if( ! __get_arm_id(card, 1, date) )   goto hell;
	cardinfo(card->cardnum, "   date: %s", date);

	if( ! __get_arm_id(card, 2, card->board.serial) ) goto hell;
	cardinfo(card->cardnum, "     id: %s", card->board.serial);

	spin_unlock_irqrestore(&card->lock, flags);
	return RET_OK;

    hell:
	spin_unlock_irqrestore(&card->lock, flags);
	cardwarn(card, "Found ARM processor, dumb firmware.");
	return RET_OK;

    limbo:
	spin_unlock_irqrestore(&card->lock, flags);
	return RET_FAIL;
} //}}}

static void arm_event(struct openpci *card, unsigned char *msg)
{ //{{{
	unsigned char mess[3];

	switch(msg[1]){
		case DSP_LOOP_OFFHOOK:
			dbgport(1,card,msg[0],"Loop OffHook");
			mess[0] = DSP_LLOOP_OFFHOOK;
			mess[1] = DSP_LOOP_OFFHOOK;
			break;
		case DSP_LOOP_ONHOOK:
			dbgport(1,card,msg[0],"Loop OnHook");
			mess[0] = DSP_LLOOP_ONHOOK;
			mess[1] = DSP_LOOP_ONHOOK;
			break;
		case DSP_LOOP_POLARITY:
			dbgport(1,card,msg[0],"Loop Polarity");
			mess[0] = DSP_LLOOP_POLARITY;
			mess[1] = DSP_LOOP_POLARITY;
			break;
		case DSP_CODEC_RING:
			dbgport(1,card,msg[0],"Ring On");
			mess[0] = DSP_LCODEC_RING;
			mess[1] = DSP_CODEC_RING;
			break;
		case DSP_RING_OFF:
			dbgport(1,card,msg[0],"Ring Off");
			mess[0] = DSP_LRING_OFF;
			mess[1] = DSP_RING_OFF;
			break;
		case DSP_CODEC_HKOFF:
		{
			struct channel *chan = &card->chans[msg[0]];
			dbgport(1,card,msg[0],"Station OffHook");
			mess[0] = DSP_LCODEC_HKOFF;
			mess[1] = DSP_CODEC_HKOFF;
			chan->state = CH_OFFHOOK;  // start audio 
			break;
		}
		case DSP_CODEC_HKON:
			dbgport(1,card,msg[0],"Station OnHook");
			mess[0] = DSP_LCODEC_HKON;
			mess[1] = DSP_CODEC_HKON;
			card->chans[msg[0]].state = CH_IDLE;
			break;
		case DSP_CODEC_FLASH:
			dbgport(1,card,msg[0],"Station Flash");
			mess[0] = DSP_LCODEC_FLASH;
			mess[1] = DSP_CODEC_FLASH;
			break;

		case DSP_PROSLIC_SANITY:
			portcrit(card,msg[0],"ProSlic has gone INSANE");
			return;
		case DSP_PROSLIC_PWR_ALARM:
		{
			char errbuf[32] = " Unknown", *p = errbuf;
			int i = 49;

			msg[2] >>= 2;
			for(; i < 55; ++i, msg[2] >>= 1 )
			    if(msg[2] & 1){ *(++p)='Q'; *(++p)=i; *(++p)=','; }
			if( p != errbuf ) *p = '\0';
			portcrit(card,msg[0],"ProSlic power ALARM:%s",errbuf);
			return;
		}
		case DSP_VDAA_ISO_FRAME_E:
			portcrit(card,msg[0],"VDAA lost ISO-Cap frame lock");
			return;

		default:
			portwarn(card,msg[0],"Unknown message [%d] from Arm", msg[1]);
			return;
	}
	mess[2] = msg[0];
	if( vt_send_event(&card->board, mess, mess[0]) )
		portcrit(card,msg[0],"FAILED to forward arm event into MSGQ");
} //}}}

/* You must hold the card spinlock to call this function */
static inline int __read_arm_byte( struct openpci *card, unsigned char *msg )
{ //{{{
	int i;

	HRXF_WAIT(); *msg = inb(PIB(1)); ++card->armbin;
	return RET_OK;
} //}}}

static inline int read_arm_msg( struct openpci *card, unsigned char *msg )
{ //{{{
	unsigned long flags;
	int i, d, count;
	int ret = RET_OK;

	spin_lock_irqsave(&card->lock, flags);
	outb(0x08, PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED();
	//XXX Do we need to clear the interrupt flag even if this fails?
	HRXF_WAIT_LOCKED(); count = inb(PIB(1)); ++card->armbin;
	if( count == 0 ){
		ret = RET_FAIL;
	} else if( count < 3 || count > 4 ){
		cardcrit(card->cardnum, "BOGUS arm message size %d, flushing queue", count);
		// NB: This may take a while (up to 500usec or more) to complete
		//     and we are in the isr at present when this is called, so
		//     we may miss an interrupt or two while this is done in the
		//     bottom half, but we are already in trouble, so...
		d = debug; debug = 5; __ping_arm( card ); debug = d;
		ret = RET_FAIL;
	} else while( --count ){
		if( ! __read_arm_byte(card, msg) ){
			cardcrit(card->cardnum,
				 "Failed to read arm message %d more bytes expected",
				 count);
			ret = RET_FAIL;
			break;
		}
		++msg;
	}
	outb(0x09, PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED();
	spin_unlock_irqrestore(&card->lock, flags);
	return ret;
} //}}}

static void openpci_arm_work( void *cardptr )
{ //{{{
	struct openpci *card = (struct openpci*)cardptr;
	unsigned char armmsg[3];

	if( read_arm_msg(card, armmsg) )  arm_event(card, armmsg);
} //}}}

/* You must hold the card spinlock to call this function */
static void __start_dma(struct openpci *card)
{ //{{{
	outb(0x0f, TJ_CNTL);       // Reset TDM bus
	udelay(100);               // Wait two cycles of the ARM slow clock
	outb(TJ_RUNFLAGS, TJ_CNTL);
	outb(0x01, TJ_OPER);
} //}}}

static void openpci_write(struct openpci *card, int status)
{ //{{{
	struct vtboard *board = &card->board;
	struct channel *chans = card->chans;
	volatile unsigned int *writeblock;
	int x,y;

	if (status & 0x01){
		writeblock = card->writeblock;

	    #if VERBOSE_TIMING
		if (status & 0x02){
			cardcrit(card->cardnum,
				 "Write buffer overrun at interrupt %u (%lld), flags = %x, current = %x:%x",
				 card->interrupt_count, card->time_tj_last_rel,
				 status, inb(TJ_DMAWC), inb(TJ_DMARC));
		}
	    #endif
	} else if( status & 0x02 ){
		writeblock = card->writeblock + FRAME*2;
	} else {
		cardcrit(card->cardnum, "bad write interrupt flags %x, at %x / %d",
			status, inb(TJ_DMAWC), card->interrupt_count);  //XXX lock?
		return;
	}
	vt_write(board);
	for(y=0,x=0; x < FRAME; ++x){
		if(chans[0].porttype) writeblock[y]  = chans[0].txbuf[x] << 24;
		else writeblock[y] = 0;
		if(chans[1].porttype) writeblock[y] |= chans[1].txbuf[x] << 16;
		if(chans[2].porttype) writeblock[y] |= chans[2].txbuf[x] << 8;
		if(chans[3].porttype) writeblock[y] |= chans[3].txbuf[x];
		++y;
		writeblock[y] = 0;
		if(chans[4].porttype) writeblock[y] |= chans[4].txbuf[x] << 24;
		// ensure first port is never 0x00
		if( ! writeblock[y] ) writeblock[y] = 0x01000000;
		if(chans[5].porttype) writeblock[y] |= chans[5].txbuf[x] << 16;
		if(chans[6].porttype) writeblock[y] |= chans[6].txbuf[x] << 8;
		if(chans[7].porttype) writeblock[y] |= chans[7].txbuf[x];
		++y;
	}
} //}}}

static void openpci_read(struct openpci *card, int status)
{ //{{{{
	struct vtboard *board = &card->board;
	struct channel *chans = card->chans;
	volatile unsigned int *readblock;
	int x,y;

	if (status & 0x08){
		readblock = card->readblock + FRAME*2;

	    #if VERBOSE_TIMING
		if (status & 0x04){
			cardcrit(card->cardnum,
				 "Read buffer overrun at interrupt %u (%lld), flags = %x, current = %x:%x",
				 card->interrupt_count, card->time_tj_last_rel,
				 status, inb(TJ_DMARC), inb(TJ_DMAWC));
		}
	    #endif
	} else if( status & 0x04 ){
		readblock = card->readblock;
	} else {
		cardcrit(card->cardnum, "bad read interrupt flags %x, at %x / %d",
			status, inb(TJ_DMARC), card->interrupt_count);	//XXX lock?
		return;
	}
	for(y=0,x=0; x < FRAME; ++x){
		if(chans[4].porttype) chans[4].rxbuf[x] = (readblock[y] >> 24) & 0xff;
		if(chans[5].porttype) chans[5].rxbuf[x] = (readblock[y] >> 16) & 0xff;
		if(chans[6].porttype) chans[6].rxbuf[x] = (readblock[y] >> 8) & 0xff;
		if(chans[7].porttype) chans[7].rxbuf[x] =  readblock[y] & 0xff;
		++y;
		if(chans[0].porttype) chans[0].rxbuf[x] = (readblock[y] >> 24) & 0xff;
		if(chans[1].porttype) chans[1].rxbuf[x] = (readblock[y] >> 16) & 0xff;
		if(chans[2].porttype) chans[2].rxbuf[x] = (readblock[y] >> 8) & 0xff;
		if(chans[3].porttype) chans[3].rxbuf[x] =  readblock[y] & 0xff;
		++y;
	}
	vt_read(board);
} //}}}

static irqreturn_t openpci_isr(int irq, void *dev_id, struct pt_regs *regs)
{ //{{{
	struct openpci *card = dev_id;
	unsigned long flags;
	unsigned char status;

    #if VERBOSE_TIMING
	unsigned long long start_isr, end_isr, time_tmp;

	rdtscll(start_isr);

	if( card->time_tj_last != 0 ){
	    time_tmp = start_isr - card->time_tj_last;
	    if( card->delay_tj_max < time_tmp ){
		card->delay_tj_max = time_tmp;
		cardinfo(card->cardnum, "tj interrupt delay time = %lld", card->delay_tj_max );
	    }
	    card->time_tj_last_rel = time_tmp;
	}
	card->time_tj_last = start_isr;
    #endif


	/* Check what interrupted us */
	spin_lock_irqsave(&card->lock, flags);
	status = inb(TJ_INTSTAT);
	outb(status, TJ_INTSTAT);

	/* No TJ status flags?  Check for an ARM interrupt */
	if (!status){

		/* Was it really our arm? */
		if (inb(TJ_AUXR) & 0x2){
			spin_unlock_irqrestore(&card->lock, flags);
			return IRQ_NONE;
		}
		spin_unlock_irqrestore(&card->lock, flags);
		++card->arm_interrupt_count;

	    #if VERBOSE_TIMING
		rdtscll( end_isr );
		time_tmp = end_isr - start_isr;
		if( time_tmp > card->time_arm_max ) {
			card->time_arm_max = time_tmp;
			cardinfo(card->cardnum, "arm interrupt service time = %lld",
						 card->time_tj_max );
		}
	    #endif
		//XXX schedule_work( &card->arm_work );
		//queue_work( openpci_workqueue, &card->arm_work );
		openpci_arm_work( card );
		return IRQ_HANDLED;
	}
	if (status & 0x10){
		/* To recover the PCI master state machine set the reset bit
		   in Reg0 to 1 and reset it to 0, also enable DMA operation
		   by setting Reg1 bit0 to 1.
		   Just starting the DMA should do all this
		 */
		__start_dma(card);
		spin_unlock_irqrestore(&card->lock, flags);

		cardcrit(card->cardnum, "PCI Master Abort.");
		return IRQ_HANDLED;
	}
	spin_unlock_irqrestore(&card->lock, flags);

	if (status & 0x20){
		/* PCI Target abort */
		cardcrit(card->cardnum, "PCI Target Abort.");
		return IRQ_HANDLED;
	}
	if (status & 0x03){
		openpci_write(card, status);
	}
	if (status & 0x0c){
		/* Only count read interrupts, we use this count for timing
		 * elsewhere, and the 'per 1ms' assumption breaks if we clock
		 * both read and write, which may arrive independently */
		++card->interrupt_count;

		/* Transfer Audio */
		/* de-multiplex to tmp buffers, need to add tmp buffers to structure some where */
		/* pass to vtcore to put into user space */
		openpci_read(card, status);

	    #ifdef DEBUG_LOOP_VOLTAGE
	    {
		int port   = card->interrupt_count & 0x07;
		int ignore = card->interrupt_count & 0xf0;

		if( ! ignore && (voltmeter & ((1 << (card->cardnum * 8)) << port)) ) {
			unsigned char lv;
			if( card->chans[port].porttype == VT_PORT_VDAA
			 && read_reg_fxo(card, port, VDAR_LOOPLV, &lv) )
				portinfo(card,port,"loop voltage %d",
						   lv < 128 ? lv : lv - 256);
		}
	    }
	    #endif
	}

    #if VERBOSE_TIMING
	rdtscll( end_isr );
	time_tmp = end_isr - start_isr;
	if( time_tmp > card->time_tj_max ){
	    card->time_tj_max = time_tmp;
	    cardinfo(card->cardnum, "tj interrupt service time = %lld", card->time_tj_max );
	}
	if (!(card->interrupt_count%60000)){
		cardinfo(card->cardnum, "delay = %lld, tjmax = %lld, armmax = %lld",
			 card->delay_tj_max, card->time_tj_max, card->time_arm_max );
	}
	// Some values get distorted due to the time taken to probe each card
	// clear them every 5 mins to get a better baseline.
	if (!(card->interrupt_count%300000)){
	    cardinfo(card->cardnum, "Clearing delay counts");
	    card->time_tj_last = 0;
	    card->time_tj_last_rel = 0;
	    card->time_tj_max = 0;
	    card->delay_tj_max = 0;
	    card->time_arm_max = 0;
	}
    #endif

	return IRQ_HANDLED;
} //}}}

static int vtopenpci_set_country(void *carddata, int port, const char *name)
{ //{{{
	struct openpci *card = carddata;

	switch( card->chans[port].porttype ){
	    case VT_PORT_VDAA:
		return configure_vdaa_country(card,port,name);

	    case VT_PORT_PROSLIC:
		return configure_proslic_country(card,port,name);
	}
	return 0;
} //}}}

static int vtopenpci_set_playgain(void *carddata, int port, int gain)
{ //{{{
	return set_play_gain( carddata, port, gain );
} //}}}

static int vtopenpci_get_playgain(void *carddata, int port, int *err)
{ //{{{
	struct openpci *card = carddata;
	signed char    maj,min;
	unsigned short gain;

	switch( card->chans[port].porttype ){
	    case VT_PORT_VDAA:
		if( read_reg_fxo(card, port, VDAR_TXGAIN2, &maj) &&
		    read_reg_fxo(card, port, VDAR_TXGAIN3, &min) )
		{
			dbgport(4,card,port,"got vdaa tx gain: %#x:%#x", maj, min);
			if( maj & (1<<4) ){
				maj &= 0xf;
				maj = -maj;
			}
			if( min & (1<<4) ){
				min &= 0x0f;
				min = -min;
			}
			*err = 0;
			return maj * 10 + min + 0x80;
		} else  *err = -EIO;
		break;

	    case VT_PORT_PROSLIC:
		// For SLIC, tx and rx are relative to the handset.
		if( read_indreg_fxs(card, port, PSI_RX_DIG_GAIN, &gain) ){
			dbgport(4,card,port,"got proslic rx gain: %#x", gain);
			*err = 0;
			return gain >> 7;
		} else  *err = -EIO;
		break;

	    default:
		*err = -EINVAL;
	}
	return 0;
} //}}}

static int vtopenpci_set_recgain(void *carddata, int port, int gain)
{ //{{{
	return set_rec_gain( carddata, port, gain );
} //}}}

static int vtopenpci_get_recgain(void *carddata, int port, int *err)
{ //{{{
	struct openpci *card = carddata;
	signed char    maj,min;
	unsigned short gain;

	switch( card->chans[port].porttype ){
	    case VT_PORT_VDAA:
		if( read_reg_fxo(card, port, VDAR_RXGAIN2, &maj) &&
		    read_reg_fxo(card, port, VDAR_RXGAIN3, &min) )
		{
			dbgport(4,card,port,"got vdaa rx gain: %#x:%#x", maj, min);
			if( maj & (1<<4) ){
				maj &= 0xf;
				maj = -maj;
			}
			if( min & (1<<4) ){
				min &= 0x0f;
				min = -min;
			}
			*err = 0;
			return maj * 10 + min + 0x80;
		} else *err = -EIO;
		break;

	    case VT_PORT_PROSLIC:
		// For SLIC, tx and rx are relative to the handset.
		if( read_indreg_fxs(card, port, PSI_TX_DIG_GAIN, &gain) ){
			dbgport(4,card,port,"got proslic tx gain: %#x", gain);
			*err = 0;
			return gain >> 7;
		} else *err = -EIO;
		break;

	    default:
		*err = -EINVAL;
	}
	return 0;
} //}}}

static int vtopenpci_set_hook(void *carddata, int port, int hookstate)
{ //{{{
	if( hookstate < CH_IDLE || hookstate > CH_OFFHOOK ) return -EINVAL;
	return set_hook_state( carddata, port, hookstate );
} //}}}

static int vtopenpci_get_hook(void *carddata, int port, int *err)
{ //{{{
	struct channel *chan = &((struct openpci*)carddata)->chans[port];

	*err = chan->porttype ? 0 : -EINVAL;
	return chan->state;
} //}}}

static int vtopenpci_set_ring(void *carddata, int port, int ringing)
{ //{{{
	struct openpci *card = carddata;
	struct channel *chan = &card->chans[port];
	unsigned long flags;
	int i;

	if( chan->porttype != VT_PORT_PROSLIC ) return -EINVAL;
	spin_lock_irqsave(&card->lock, flags);
	outb(0x21,PIB(1));    ++card->armbout; HTXF_WAIT_LOCKED_RET(-EIO);
	outb(port,PIB(1));    ++card->armbout; HTXF_WAIT_LOCKED_RET(-EIO);
	outb(ringing,PIB(1)); ++card->armbout; HTXF_WAIT_LOCKED_RET(-EIO);
	spin_unlock_irqrestore(&card->lock, flags);
	return 0;
} //}}}

static int do_get_ring(struct openpci *card, int port, int *value)
{ //{{{
	unsigned long flags;
	int i;

	spin_lock_irqsave(&card->lock, flags);
	outb(0x22,PIB(1));  ++card->armbout; HTXF_WAIT_LOCKED();
	outb(port,PIB(1));  ++card->armbout; HTXF_WAIT_LOCKED();
	HRXF_WAIT_LOCKED(); ++card->armbin; *value = inb(PIB(1));
	spin_unlock_irqrestore(&card->lock, flags);

	return RET_OK;
} //}}}

static int vtopenpci_get_ring(void *carddata, int port, int *err)
{ //{{{
	struct openpci *card = carddata;
	struct channel *chan = &card->chans[port];
	int value = 0;

	*err = chan->porttype ? 0 : -EINVAL;
	if( ! do_get_ring( card, port, &value ) ) *err = -EIO;
	return value;
} //}}}

static void get_module_info( struct openpci *card )
{ //{{{
	unsigned long flags;
	char buf[VT_MOD_INFO_SIZE + 1] = { 0 };
	int i = 0;

	spin_lock_irqsave(&card->lock, flags);
	for(; i < 2; ++i){
		struct modinfo *m = &card->moduleinfo[i];

		if( ! __get_module_info( card, i, 0, m->type ) )   goto hell;
		if( ! __get_module_info( card, i, 1, buf ) )       goto hell;
		m->ports = *buf;
		if( ! __get_module_info( card, i, 2, buf ) )       goto hell;
		m->rev = *buf;
		if( ! __get_module_info( card, i, 3, m->serial ) ) goto hell;
		if( ! __get_module_info( card, i, 4, m->date ) )   goto hell;

		if( m->type[0] == 'F' ){
			cardinfo(card->cardnum, "module %d: %d port %s, revision %d",
						i, m->ports, m->type, m->rev);
			cardinfo(card->cardnum, "  serial: %s", m->serial);
			cardinfo(card->cardnum, "    date: %s", m->date);
		} else  cardinfo(card->cardnum, "No module %d", i);
	}
	spin_unlock_irqrestore(&card->lock, flags);
	return;

    hell:
	spin_unlock_irqrestore(&card->lock, flags);
	cardinfo(card->cardnum, "failed to get module info");
} //}}}

static int create_module_proc_const_int(struct openpci *card, int module,
					const char *node, long val)
{ //{{{
	struct proc_dir_entry *p;

	p = create_proc_read_entry(node, 0444, card->moduleinfo[module].procfs_root,
					 vt_proc_read_int, (void*)val);
	if( p ){
		p->owner = THIS_MODULE;
		return RET_OK;
	}
	crit("failed to create procfs module%d/%s", module, node);
	return RET_FAIL;
} //}}}

static int create_module_proc_const_string(struct openpci *card, int module,
					   const char *node, const char *str)
{ //{{{
	struct proc_dir_entry *p;

	p = create_proc_read_entry(node, 0444, card->moduleinfo[module].procfs_root,
					 vt_proc_read_string, (void*)str);
	if( p ){
		p->owner = THIS_MODULE;
		return RET_OK;
	}
	crit("failed to create procfs module%d/%s", module, node);
	return RET_FAIL;
} //}}}

// The card must be registered with vtcore before this function is called.
static void export_module_info( struct openpci *card )
{ //{{{
	int i = 0;

	if( ! card->board.procfs_root ) return;
	for(; i < 2; ++i){
		struct modinfo *mod = &card->moduleinfo[i];
		struct proc_dir_entry **modroot = &mod->procfs_root;

		if( mod->type[0] != 'F' ) continue;
		if( ! *modroot ) {
			char  dirname[16] = { 0 };

			snprintf(dirname, sizeof(dirname)-1, "module%d", i);
			*modroot = proc_mkdir(dirname, card->board.procfs_root);
		}
		if( ! *modroot ){
			cardcrit(card->cardnum, "failed to create procfs module%d", i);
			continue;
		}
		(*modroot)->owner = THIS_MODULE;

		create_module_proc_const_string(card, i, "type", mod->type );
		create_module_proc_const_int(card, i, "ports", mod->ports );
		create_module_proc_const_int(card, i, "revision", mod->rev );
		create_module_proc_const_string(card, i, "id", mod->serial );
		create_module_proc_const_string(card, i, "date", mod->date );
	}
} //}}}

static void unexport_module_info( struct openpci *card )
{ //{{{
	int i = 0;

	if( ! card->board.procfs_root ) return;
	for(; i < 2; ++i){
		char  dirname[16] = { 0 };
		struct proc_dir_entry *modroot = card->moduleinfo[i].procfs_root;

		snprintf(dirname, sizeof(dirname)-1, "module%d", i);

		remove_proc_entry("type", modroot);
		remove_proc_entry("ports", modroot);
		remove_proc_entry("revision", modroot);
		remove_proc_entry("id", modroot);
		remove_proc_entry("date", modroot);
		remove_proc_entry(dirname, card->board.procfs_root);
	}
} //}}}

static int vtopenpci_board_register( struct openpci *card )
{ //{{{
	struct vtboard *board = &card->board;

	/* Register with vtcore */
	board->owner        = THIS_MODULE;
	board->drvdata      = card;
	board->name         = vtopenpci;
	board->maxports     = MAX_PORTS;
	board->chans        = card->chans;
	board->ioctl        = vtopenpci_ioctl;
	board->set_country  = vtopenpci_set_country;
	board->set_playgain = vtopenpci_set_playgain;
	board->get_playgain = vtopenpci_get_playgain;
	board->set_recgain  = vtopenpci_set_recgain;
	board->get_recgain  = vtopenpci_get_recgain;
	board->set_hook     = vtopenpci_set_hook;
	board->get_hook     = vtopenpci_get_hook;
	board->set_ring     = vtopenpci_set_ring;
	board->get_ring     = vtopenpci_get_ring;

	if( ! vt_board_register(board) ) return RET_FAIL;
	dbginfo(2,card->cardnum, "registered for %d ports with vtcore",
				  board->maxports);

	vt_create_board_proc_const_int(board, "firmware", card->firmware);

	return RET_OK;
} //}}}

static void vtopenpci_board_unregister( struct openpci *card )
{ //{{{
	struct vtboard *board = &card->board;

	remove_proc_entry("firmware", board->procfs_root);
	vt_board_unregister(board);
} //}}}

static int __devinit openpci_probe_board(struct pci_dev *pdev,
					 const struct pci_device_id *ent)
{ //{{{
	struct openpci *card;
	int cardnum = 0;
	int failret = -ENOMEM;
	unsigned long flags;

	if( ent->driver_data != (kernel_ulong_t)&vtopenpci )
	{
		info("Probe of non-OpenPCI card, ignoring.");
		return -EINVAL;
	}
	card = kzalloc(sizeof(struct openpci), GFP_KERNEL);
	if(!card) {
		crit("FAILED to allocate memory for new card");
		return -ENOMEM;
	}

	mutex_lock(&cards_mutex);
	for(; cardnum < MAX_CARDS && cards[cardnum]; ++cardnum);
	if (cardnum >= MAX_CARDS){
		crit("Too many OpenPCI cards(%d), max is %d.", cardnum, MAX_CARDS);
		mutex_unlock(&cards_mutex);
		goto hell;
	}
	cards[cardnum]=card;
	mutex_unlock(&cards_mutex);

	spin_lock_init(&card->lock);
	pci_set_drvdata(pdev,card);

	card->cardnum = cardnum;
	card->dev     = pdev;
	card->armbout = 0;

    #if VERBOSE_TIMING
	card->time_tj_last     = 0;
	card->time_tj_last_rel = 0;
	card->time_tj_max      = 0;
	card->delay_tj_max     = 0;
	card->time_arm_max     = 0;
    #endif

	dbginfo(1,cardnum, "initialising card");
	if (pci_enable_device(pdev)){
		failret = -EIO;
		goto hell_2;
	}
	//XXX should we call pci_disable_device if we fail beyond this point.
	//    yes, but not if request_region fails, meaning some other driver
	//    already has the card.  This shouldn't usually happen though.

	card->ioaddr = pci_resource_start(pdev,0);
	if( ! request_region(card->ioaddr, pci_resource_len(pdev,0), NAME) ){
		cardcrit(cardnum, "FAILED to lock IO region, another driver using it?");
		failret = -EBUSY;
		goto hell_2;
	}

	spin_lock_irqsave(&card->lock, flags);
	outb(0xff, TJ_AUXD);            /* Set up TJ to access the ARM */
	outb(0x78, TJ_AUXC);            /* Set up for Jtag */
	outb(0x00, TJ_CNTL);            /* pull ERST low */
	spin_unlock_irqrestore(&card->lock, flags);
	msleep(1);	                /* Wait a bit */

	dbginfo(1,cardnum,"Starting ARM");
	spin_lock_irqsave(&card->lock, flags);
	outb(TJ_RUNFLAGS, TJ_CNTL);     /* pull ERST high again */
	spin_unlock_irqrestore(&card->lock, flags);
	msleep(100);                    /* Give it all a chance to boot */

	if(program == NULL){
		if( ! check_arm(card) ){
			cardcrit(cardnum, "FAILED to find ARM processor");
			failret = -EIO;
			goto hell_3;
		}
		if( card->firmware < 9 ){
			cardcrit(cardnum,
				 "Firmware version %d not supported by this driver",
				 card->firmware);
			cardcrit(cardnum, " contact VoiceTronix to have it updated");
			failret = -ENODEV;
			goto hell_3;
		}
		if( ! check_ports(card) ){
			cardcrit(cardnum, "FAILED to find any ports");
			failret = -EIO;
			goto hell_3;
		}

		card->writeblock = (int *)pci_alloc_consistent(pdev,
							       VT_PCIDMA_BLOCKSIZE,
							       &card->writedma);
		if (!card->writeblock){
			cardcrit(cardnum, "FAILED to get DMA memory.");
			goto hell_3;
		}
		card->readblock = card->writeblock + FRAME * (MAX_PORTS*2 / sizeof(int));
		card->readdma = card->writedma + FRAME * (MAX_PORTS*2);

		/* Initialise the DMA area */
		memset((void*)card->writeblock, 0x55, VT_PCIDMA_BLOCKSIZE);

		spin_lock_irqsave(&card->lock, flags);
		/* Configure the TDM bus for MSB->LSB operation */
		outb(0xc1, TJ_SERCTL);

		/* Set FSC delay to 0 so it's properly aligned */
		outb(0x0, TJ_FSCDELAY);

		/* Setup DMA access */
		outl(card->writedma,                    TJ_DMAWS); /* Start */
		outl(card->writedma + VT_PCIDMA_MIDDLE, TJ_DMAWI); /* Middle */
		outl(card->writedma + VT_PCIDMA_END,    TJ_DMAWE); /* End */
		outl(card->readdma,                     TJ_DMARS); /* Start */
		outl(card->readdma + VT_PCIDMA_MIDDLE,  TJ_DMARI); /* Middle */
		outl(card->readdma + VT_PCIDMA_END,     TJ_DMARE); /* End */

		/* Clear interrupts */
		outb(0xff, TJ_INTSTAT);
		spin_unlock_irqrestore(&card->lock, flags);

		//XXX Can we enable this by default in the ARM now?
		if( ! arm_monitor(card,1) ){
			cardcrit(card->cardnum, "FAILED to start arm monitoring");
			goto hell_4;
		}

		if( ! configure_ports(card) ){
			cardcrit(cardnum, "FAILED to configure ports.");
			goto hell_4;
		}

		if( ! vtopenpci_board_register( card ) ) {
			cardcrit(cardnum, "FAILED to register with vtcore");
			failret = -EIO;
			goto hell_4;
		}

		/* Enable bus mastering */
		pci_set_master(pdev);

		/* we must be ready to handle requests before this point */
		if (request_irq(pdev->irq, openpci_isr, shareirq ? SA_SHIRQ : 0, NAME, card)){
			cardcrit(cardnum, "FAILED to get IRQ");
			failret = -EIO;
			goto hell_5;
		}
		dbginfo(1,cardnum, "using IRQ %d", pdev->irq);

		//XXX Is this the right place for these?
		//INIT_WORK( &card->arm_work, openpci_arm_work, card );

		/* Enable interrupts */
		spin_lock_irqsave(&card->lock, flags);
		outb(0x3f, TJ_MASK0);
	  //XXX outb(0x0,  TJ_AUXINTPOL);
		outb(0x02, TJ_MASK1);

		__start_dma(card);
		spin_unlock_irqrestore(&card->lock, flags);
	}
	else {
		/* We going into program card mode! */
		if(!strcmp(program,"cpld")){
			card->ppdatamask = 0x8;
			cardinfo(cardnum, "Setting card into CPLD program mode");
		} else {
			if( check_arm(card) ) get_module_info(card);
			else cardinfo(cardnum, "failed to boot ARM processor");
			card->ppdatamask = 0x0;
			cardinfo(cardnum, "Setting card into ARM program mode");
		}
		if( ! vtopenpci_board_register( card ) ){
			cardcrit(cardnum, "FAILED to register with vtcore");
			failret = -EIO;
			goto hell_3;
		}
		if( card->ppdatamask == 0 ) export_module_info(card);
	}
	dbginfo(1,cardnum,"card initialised");

	return 0;

    hell_5:
	vtopenpci_board_unregister( card );
    hell_4:
	if (card->writeblock){
		pci_free_consistent(pdev, VT_PCIDMA_BLOCKSIZE,
				    (void*)card->writeblock, card->writedma);
	}
    hell_3:
	outb(0x00, TJ_CNTL);
	release_region(card->ioaddr, pci_resource_len(pdev,0));
    hell_2:
	cards[cardnum] = NULL;
    hell:
	kfree(card);
	return failret;
} //}}}

// Clearly this can't really put a station on-hook, but since we're about
// to unplug it from our side anyway, that soon will be moot.  This signals
// all our dependents to consider the station inactive.
static void force_station_onhook(struct openpci *card, int port)
{ //{{{
	unsigned char mess[] = {DSP_LCODEC_HKON, DSP_CODEC_HKON, port};

	if( card->chans[port].state != CH_IDLE ){
		dbgport(1,card,port,"forcing station from state %d to idle",
			card->chans[port].state );
		card->chans[port].state = CH_IDLE;
		vt_send_event(&card->board, mess, mess[0]);
	}
} //}}}

static void __devexit openpci_remove_board(struct pci_dev *pdev)
{ //{{{
	struct openpci *card = pci_get_drvdata(pdev);
	unsigned long flags;
	int i=0, n;

	if (!card)
		return;

	n = card->cardnum;
	dbginfo(1,n, "releasing card");
	if(program == NULL){
		spin_lock_irqsave(&card->lock, flags);
		outb(0x07, PIB(1) );                  /* Stop arm monitoring */
		//XXX ++card->armbout; HTXF_WAIT_LOCKED();
		outb(0x00, TJ_OPER);                  /* Stop DMA access */
		outb(0x00, TJ_MASK0);                 /* Disable interrupts */
		outb(0x00, TJ_MASK1);
		spin_unlock_irqrestore(&card->lock, flags);

		free_irq(pdev->irq, card);

		// Hangup on anyone still offhook, vtcore will refuse
		// to unregister a channel that is not already idle.
		for(; i < MAX_PORTS; ++i) {
			if( card->chans[i].state != CH_IDLE ) {
				switch( card->chans[i].porttype ){
				    case VT_PORT_VDAA:
					set_hook_state(card, i, HOOK_ONHOOK);
					break;
				    case VT_PORT_PROSLIC:
					force_station_onhook(card, i);
					break;
				    default:
					portcrit(card,i,
						 "%s in state %d not idle?",
						 porttype(card,i),
						 card->chans[i].state );
				}
			}
		}
		vtopenpci_board_unregister( card );
		outb(0x00, TJ_CNTL);

		pci_free_consistent(pdev, VT_PCIDMA_BLOCKSIZE,
				    (void*)card->writeblock, card->writedma);
	} else {
		//XXX Do we need to do a forceful hangup here too?
		if( card->ppdatamask == 0 ) unexport_module_info(card);
		vtopenpci_board_unregister( card );
		outb(0x00, TJ_CNTL);
	}
	release_region(card->ioaddr, pci_resource_len(pdev,0));

	dbginfo(1,n, "Serviced %d interrupts",card->interrupt_count);
	dbginfo(1,n, "Serviced %d Arm interrupts",card->arm_interrupt_count);
	dbginfo(1,n, "Sent %d bytes to the Arm",card->armbout);
	dbginfo(1,n, "Received %d bytes from the Arm",card->armbin);

	mutex_lock(&cards_mutex);
	cards[n] = NULL;
	mutex_unlock(&cards_mutex);

	kfree(card);
	cardinfo(n, "card released");
} //}}}


module_init(vtopenpci_init);
module_exit(vtopenpci_exit);

MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_VERSION(VT_VERSION);
MODULE_LICENSE("GPL");


//XXX
#if 0
//{{{
static char dtmf_lookup[] = "1234567890*#ABCB";
static void check_dtmf(struct openpci *card, int port)
{ //{{{
	unsigned char digit;
	char mess[DSP_LDTMF];

	if( read_reg_fxs(card, port, PS_DTMF_OUT, &digit) ){
		mess[0] = DSP_LDTMF;
		mess[1] = DSP_DTMF;
		mess[2] = port;
		mess[3] = dtmf_lookup[digit&0x0F];
		if( vt_send_event(&card->board, mess, DSP_LDTMF) )
			cardinfo(card->cardnum,
				 "%02d: failed to send DSP_LDTMF event into MSGQ", port);
	} else
		portcrit(card,port,"check_dtmf failed to read PS_DTMF_OUT");
} //}}}

// Do not call this from an interrupt context, it may sleep.
static void reset_arm(struct openpci *card)
{ //{{{
	cardinfo(card->cardnum, "Reseting ARM");

	/* Flush Byte */
	outb(0x00, PIB(1)); ++card->armbout;
	/* Stop the ARM chip*/
	outb(0x00, TJ_CNTL);
	/* Wait a bit */
	msleep(50);

	/* Flush Byte */
	outb(0x00, PIB(1)); ++card->armbout;
	/* Start the TJ and ARM chip*/
	outb(TJ_RUNFLAGS, TJ_CNTL);
	/* Wait a bit more */
	msleep(450);
} //}}}

// Do not call this from an interrupt context, it may sleep.
static int flush_arm(struct openpci *card)
{ //{{{
	//XXX Er, what is going on here?  Why do this, maybe once,
	//    then definitely 100 times??
	//    Nevermind for now, use ping_arm to flush it instead.
	int i,x = 100;
	if(!HTXF_READY){
		outb(0x00, PIB(1)); ++card->armbout; HTXF_WAIT();
	}
	while(--x){
		outb(0x00, PIB(1)); ++card->armbout; HTXF_WAIT();
	}
	msleep(500);
	return RET_OK;
} //}}}

//}}}
#endif

